from ..model_execution.bin import create_binary_files, create_binary_tar_file, check_conversion_tools, \
    check_inputs_directory, cleanup_bin_directory
from .cleaners import PathCleaner
from .base import OasisBaseCommand


class BuildCmd(OasisBaseCommand):
    """
    Builds input binary files for model execution.
    """

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(BuildCmd, self).add_args(parser)

        parser.add_argument(
            'source', default='.', type=PathCleaner('Source', preexists=True), nargs='?',
            help='The csv source directory.'
        )

        parser.add_argument(
            'destination', default=None, type=PathCleaner('Destination', preexists=False), nargs='?',
            help='The binary destination directory, by default this is the same as the source.'
        )

        parser.add_argument(
            '--do-il', action='store_true',
            help='Flag to build insured loss calculation binaries.'
        )

        parser.add_argument(
            '--build-tar', action='store_true',
            help='Flag to build the binary tar.'
        )

    def action(self, args):
        """
        Builds the input binary files

        :param args: The arguments from the command line
        :type args: Namespace
        """
        destination = args.destination or args.source

        check_conversion_tools(do_il=args.do_il)
        check_inputs_directory(args.source, do_il=args.do_il, check_binaries=False)
        create_binary_files(args.source, destination, do_il=args.do_il)

        if args.build_tar:
            create_binary_tar_file(destination)


class CleanCmd(OasisBaseCommand):
    """
    Cleans up all binary files.
    """

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(CleanCmd, self).add_args(parser)

        parser.add_argument(
            'target', default='.', type=PathCleaner('Target'), nargs='?',
            help='The directory to clean.'
        )

    def action(self, args):
        """
        Deletes all input binary files

        :param args: The arguments from the command line
        :type args: Namespace
        """
        cleanup_bin_directory(args.target)


class CheckCmd(OasisBaseCommand):
    """
    Checks the required conversion tools and input files are present to build the input binaries.
    """

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(CheckCmd, self).add_args(parser)

        parser.add_argument(
            'target', default='.', type=PathCleaner('Target'), nargs='?',
            help='The directory to clean.'
        )

        parser.add_argument(
            '--do-il', action='store_true',
            help='Flag to check insured loss calculation inputs.'
        )

        parser.add_argument(
            '--check-binaries', action='store_true',
            help='Flag to check if binary files exist.'
        )

    def action(self, args):
        """
        Checks the required conversion tools and input files are present to build the input binaries.

        :param args: The arguments from the command line
        :type args: Namespace
        """
        check_inputs_directory(args.target, do_il=args.do_il, check_binaries=args.check_binaries)
        check_conversion_tools(do_il=args.do_il)


class BinCmd(OasisBaseCommand):
    """
    Build, clean and check input binary files
    """

    sub_commands = {
        'build': BuildCmd,
        'clean': CleanCmd,
        'check': CheckCmd,
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Python utilities used for setting up the resources needed to complete a
    model run, i.e. generating ktools outputs from Oasis files.
"""

from __future__ import print_function

import glob
import logging
import tarfile
from itertools import chain

import shutilwhich
import six
from pathlib2 import Path
from six import itervalues

__all__ = [
    'create_binary_files',
    'prepare_model_run_directory',
    'prepare_model_run_inputs'
]

import os
import shutil
import subprocess

from ..utils.exceptions import OasisException
from .files import TAR_FILE, INPUT_FILES, GUL_INPUT_FILES, IL_INPUT_FILES


def prepare_model_run_directory(
    run_dir_path,
    oasis_files_src_path=None,
    analysis_settings_json_src_file_path=None,
    model_data_src_path=None,
    inputs_archive=None,
):
    """
    Ensures that the model run directory has the correct folder structure in
    order for the model run script (ktools) to be executed.

    ::

        <run_directory>
        ├── fifo
        ├── input
        │   └── csv
        ├── output
        ├── static
        └── work

    If any subfolders are missing they are created.

    Optionally, if the path to a set of Oasis files is provided then they
    are copied into the ``input/csv`` subfolder.

    Optionally, if the path to the analysis settings JSON file is provided
    then it is copied to the base of the run directory.

    Optionally, if the path to model data is provided then the files are
    symlinked into the ``static`` subfolder provided the OS is of type
    Darwin or Linux, otherwise the source folder tree is recursively
    copied into the ``static`` subfolder.

    :param run_directory: the model run directory
    :type run_directory: str

    :param oasis_files_src_path: path to a set of Oasis files
    :type oasis_files_src_path: str

    :param analysis_settings_json_src_file_path: analysis settings JSON file path
    :type analysis_settings_json_src_file_path: str

    :param model_data_src_path: model data source path
    :type model_data_src_path: str

    :param inputs_archive: path to a tar file containing input files
    :type inputs_archive: str
    """
    try:
        for subdir in ['fifo', 'output', 'static', 'work']:
            Path(run_dir_path, subdir).mkdir(parents=True, exist_ok=True)

        if not inputs_archive:
            Path(run_dir_path, 'input', 'csv').mkdir(parents=True, exist_ok=True)
        else:
            with tarfile.open(inputs_archive) as input_tarfile:
                input_tarfile.extractall(path=(os.path.join(run_dir_path, 'input')))

        if oasis_files_src_path:
            oasis_files_destpath = os.path.join(run_dir_path, 'input', 'csv')
            for p in os.listdir(oasis_files_src_path):
                shutil.copy2(os.path.join(oasis_files_src_path, p), oasis_files_destpath)

        if analysis_settings_json_src_file_path:
            analysis_settings_json_dest_file_path = os.path.join(run_dir_path, 'analysis_settings.json')
            shutil.copyfile(analysis_settings_json_src_file_path, analysis_settings_json_dest_file_path)

        if model_data_src_path:
            model_data_dest_path = os.path.join(run_dir_path, 'static')

            for path in glob.glob(os.path.join(model_data_src_path, '*')):
                filename = os.path.basename(path)
                try:
                    os.symlink(path, os.path.join(model_data_dest_path, filename))
                except Exception:
                    shutil.copytree(model_data_src_path, os.path.join(model_data_dest_path, filename))

    except OSError as e:
        raise OasisException(e)


def _prepare_input_bin(run_directory, bin_name, model_settings, setting_key=None):
    bin_file_path = os.path.join(run_directory, 'input', '{}.bin'.format(bin_name))
    if not os.path.exists(bin_file_path):
        setting_val = model_settings.get(setting_key)

        if not setting_val:
            model_data_bin_file_path = os.path.join(run_directory, 'static', '{}.bin'.format(bin_name))
        else:
            # Format for data file names
            setting_val = setting_val.replace(' ', '_').lower()
            model_data_bin_file_path = os.path.join(run_directory, 'static', '{}_{}.bin'.format(bin_name, setting_val))

        if not os.path.exists(model_data_bin_file_path):
            raise OasisException('Could not find {} data file: {}'.format(bin_name, model_data_bin_file_path))

        shutil.copyfile(model_data_bin_file_path, bin_file_path)


def prepare_model_run_inputs(analysis_settings, run_directory):
    """
    Sets up binary files in the model inputs directory.

    :param analysis_settings: model analysis settings dict
    :type analysis_settings: dict

    :param run_directory: model run directory
    :type run_directory: str
    """
    try:
        model_settings = analysis_settings.get('model_settings', {})

        _prepare_input_bin(run_directory, 'events', model_settings, setting_key='event_set')
        _prepare_input_bin(run_directory, 'returnperiods', model_settings)
        _prepare_input_bin(run_directory, 'occurrence', model_settings, setting_key='event_occurrence_id')

        if os.path.exists(os.path.join(run_directory, 'static', 'periods.bin')):
            _prepare_input_bin(run_directory, 'periods', model_settings)
    except (OSError, IOError) as e:
        raise OasisException(e)


def check_inputs_directory(directory_to_check, do_il=False, check_binaries=True):
    """
    Check that all the required csv files are present in the directory.
    Args:
        ``directory`` (string): the directory containing the CSV files.
        ``do_il`` (bool): do insured loss. If True, FM file must be present.
    Returns:
        None
    """
    file_path = os.path.join(directory_to_check, TAR_FILE)
    if os.path.exists(file_path):
        raise OasisException("Inputs tar file already exists: {}".format(file_path))

    if do_il:
        input_files = (f['name'] for f in six.itervalues(INPUT_FILES) if f['type'] != 'optional')
    else:
        input_files = (f['name'] for f in six.itervalues(INPUT_FILES) if f['type'] not in ['optional', 'il'])

    for input_file in input_files:
        file_path = os.path.join(directory_to_check, input_file + ".csv")
        if not os.path.exists(file_path):
            raise OasisException("Failed to find {}".format(file_path))

        if check_binaries:
            file_path = os.path.join(directory_to_check, input_file + ".bin")
            if os.path.exists(file_path):
                raise OasisException("Binary file already exists: {}".format(file_path))


def create_binary_files(csv_directory, bin_directory, do_il=False):
    """
    Create the binary files.

    :param csv_directory: the directory containing the CSV files
    :type csv_directory: str

    :param bin_directory: the directory to write the binary files
    :type bin_directory: str

    :param do_il: whether to perform insured loss (IL) calculations; if true, FM file must be present
    :type do_il: bool
    """
    csvdir = os.path.abspath(csv_directory)
    bindir = os.path.abspath(bin_directory)

    if do_il:
        input_files = itervalues(INPUT_FILES)
    else:
        input_files = (f for f in itervalues(INPUT_FILES) if f['type'] != 'il')

    for input_file in input_files:
        conversion_tool = input_file['conversion_tool']
        input_file_path = os.path.join(csvdir, '{}.csv'.format(input_file['name']))
        if not os.path.exists(input_file_path):
            continue

        output_file_path = os.path.join(bindir, '{}.bin'.format(input_file['name']))
        cmd_str = "{} < {} > {}".format(conversion_tool, input_file_path, output_file_path)

        try:
            subprocess.check_call(cmd_str, stderr=subprocess.STDOUT, shell=True)
        except subprocess.CalledProcessError as e:
            raise OasisException(e)


def check_binary_tar_file(tar_file_path, check_il=False):
    """
    Checks that all required files are present

    :param tar_file_path: Path to the tar file to check
    :type tar_file_path: str

    :param check_il: Flag whether to check insured loss files
    :type check_il: bool

    :raises OasisException: If a required file is missing

    :return: True if all required files are present
    """
    expected_members = ('{}.bin'.format(f['name']) for f in six.itervalues(GUL_INPUT_FILES))

    if check_il:
        expected_members = chain(expected_members, ('{}.bin'.format(f['name']) for f in six.itervalues(IL_INPUT_FILES)))

    with tarfile.open(tar_file_path) as tar:
        for member in expected_members:
            try:
                tar.getmember(member)
            except KeyError:
                raise OasisException('{} is missing from the tar file {}.'.format(member, tar_file_path))

    return True


def create_binary_tar_file(directory):
    """
    Package the binaries in a gzipped tar.
    """
    original_cwd = os.getcwd()
    os.chdir(directory)

    with tarfile.open(TAR_FILE, "w:gz") as tar:
        for file in glob.glob('*.bin'):
            tar.add(file)

    os.chdir(original_cwd)


def check_conversion_tools(do_il=False):
    # Check that the conversion tools are available
    if do_il:
        input_files = six.itervalues(INPUT_FILES)
    else:
        input_files = (f for f in six.itervalues(INPUT_FILES) if f['type'] != 'il')

    for input_file in input_files:
        tool = input_file['conversion_tool']
        if shutilwhich.which(tool) is None:
            error_message = "Failed to find conversion tool: {}".format(tool)
            logging.error(error_message)
            raise OasisException(error_message)

    return True


def cleanup_bin_directory(directory):
    """
    Clean the tar and binary files.
    """
    for file in chain([TAR_FILE], (f + '.bin' for f in six.iterkeys(INPUT_FILES))):
        file_path = os.path.join(directory, file)
        if os.path.exists(file_path):
            os.remove(file_path)



# -*- coding: utf-8 -*-

"""
Utils for running SQL commands against a database.
"""

__all__ = [
    'bcp',
    'check_connection',
    'CONN_STRING',
    'DB_CONFIG',
    'execute',
    'fetch_one',
    'fetchall',
    'read_db_config'
]

import logging
import os

import pyodbc

from .log import oasis_log

DB_CONFIG = {}
CONN_STRING = ""


def bcp(table, outfile):
    """
    BCP a table to a file.
    """
    command = "freebcp {}.dbo.{} out {} -c -t, -U {} -P {} -S {}:{}".format(
        DB_CONFIG['database'], table, outfile,
        DB_CONFIG['username'], DB_CONFIG['password'],
        DB_CONFIG['server'], DB_CONFIG['port'])
    os.system(command)


def read_db_config(config_parser):
    """
    Read an Oasis standard db config
    """

    global DB_CONFIG, CONN_STRING

    DB_CONFIG['server'] = config_parser['FLAMINGO_DB_IP']
    DB_CONFIG['port'] = config_parser['FLAMINGO_DB_PORT']
    DB_CONFIG['username'] = config_parser['FLAMINGO_DB_USERNAME']
    DB_CONFIG['password'] = config_parser['FLAMINGO_DB_PASSWORD']
    DB_CONFIG['database'] = config_parser['FLAMINGO_DB_NAME']

    CONN_STRING = "DRIVER={};PORT={};SERVER={};DATABASE={};uid={};pwd={}".format(
        '{FreeTDS}',
        DB_CONFIG['port'], DB_CONFIG['server'], DB_CONFIG['database'],
        DB_CONFIG['username'], DB_CONFIG['password'])


@oasis_log()
def execute(sql, *parameters):
    """
    Execute a SQL statement with specified parameters.
    """
    conn = pyodbc.connect(CONN_STRING)
    conn.autocommit = True
    cursor = conn.cursor()
    cursor.execute(sql, parameters)
    conn.commit()


@oasis_log()
def fetch_one(sql, *parameters):
    """
    Execute a SQL statement with specified parameters, and return a
    single row.
    """
    conn = pyodbc.connect(CONN_STRING)
    conn.autocommit = True
    cursor = conn.cursor()
    cursor.execute(sql, parameters)
    row = cursor.fetchone()
    conn.commit()
    return row


@oasis_log()
def fetchall(sql, *parameters):
    """
    Execute a SQL statement with specified parameters, and return
    all rows.
    """
    conn = pyodbc.connect(CONN_STRING)
    conn.autocommit = True
    cursor = conn.cursor()
    cursor.execute(sql, parameters)
    rows = cursor.fetchall()
    logging.getLogger().info("Feteched {} rows".format(len(rows)))
    conn.commit()
    return rows


@oasis_log()
def check_connection():
    """
    Run a simple query against the Flamingo database
    """
    try:
        db_summary = "PORT={};SERVER={};DATABASE={}".format(
            DB_CONFIG['port'], DB_CONFIG['server'], DB_CONFIG['database'])
        logging.getLogger().info(
            "Checking connection: {}".format(db_summary))
        fetchall("SELECT * FROM version")
    except Exception as e:
        logging.getLogger().error(
            "Failed to connect to database: {}".format(db_summary))
        logging.getLogger().error(e)
        return False
    return True



# -*- coding: utf-8 -*-

__all__ = [
    'Translator'
]

import json
import logging
import os

import pandas as pd

from lxml import etree


class Translator(object):
    def __init__(self, input_path, output_path, xsd_path, xslt_path, append_row_nums=False, chunk_size=5000, logger=None):
        self.logger = logger or logging.getLogger()
        self.xsd = etree.parse(xsd_path)    # validation file
        self.xslt = etree.parse(xslt_path)  # transform file
        self.fpath_input = input_path       # file_in.csv
        self.fpath_output = output_path     # file_out.csv

        self.row_nums = append_row_nums  # Add 'ROW_ID' field to output
        self.row_limit = chunk_size      # MAX number of CSV input rows
        self.row_header_in = None        # CSV col header
        self.row_header_out = None       # CSV col header Post Transform

    def __call__(self):
        csv_reader = pd.read_csv(self.fpath_input, iterator=True, dtype=object, encoding='utf-8')
        for data, first_row_number, last_row_number in self.next_file_slice(csv_reader):
            self.logger.debug('--- lines[%d .. %d] ---', first_row_number, last_row_number)

            # Convert CSV -> XML
            xml_input_slice = self.csv_to_xml(
                self.row_header_in,
                data
            )
            self.print_xml(xml_input_slice)

            # Transform
            xml_output = self.xml_transform(xml_input_slice, self.xslt)
            self.print_xml(xml_output)

            # Validate Output
            self.logger.debug(self.xml_validate(xml_output, self.xsd))

            # Convert transform XML back to CSV
            self.xml_to_csv(
                xml_output,        # XML etree
                first_row_number,  # First Row in this slice
                last_row_number    # Last Row in this slice
            )

# --- Transform Functions ----------------------------------------------------#
    # https://pymotw.com/2/xml/etree/ElementTree/create.html
    # http://lxml.de/api/lxml.etree._Element-class.html

    def csv_to_xml(self, csv_header, csv_data):
        root = etree.Element('root')
        # fetch each row
        for row in csv_data:
            # Create new 'empty' record
            rec = etree.SubElement(root, 'rec')
            # Iter over columns and set attributs
            for i in range(0, len(row)):
                rec.set(self.row_header_in[i], row[i])
        return root

    def xml_to_csv(self, xml_elementTree, row_first, row_last):
        root = xml_elementTree.getroot()
        # Set output col headers
        if not (self.row_header_out):
            self.row_header_out = root[0].keys()

        # create Dataframe from xml and append each row
        rows = []
        for rec in root:
            rows.append(dict(rec.attrib))

        df_out = pd.DataFrame(rows, columns=self.row_header_out)

        # Add column for row_nums if set
        if self.row_nums:
            start = row_first + 1
            end = start + len(df_out)
            df_out.insert(0, 'ROW_ID', pd.Series(range(start,end)))

        # Append to output file
        if (row_first == 0):
            self.write_file_header(df_out.columns.tolist())
        df_out.to_csv(
            self.fpath_output,
            mode='a',
            encoding='utf-8',
            header=False,
            index=False,
        )

    # http://lxml.de/2.0/validation.html
    def xml_validate(self, xml_etree, xsd_etree):
        xmlSchema = etree.XMLSchema(xsd_etree)
        self.print_xml(xml_etree)
        self.print_xml(xsd_etree)
        # Calling 'assertValid' will raise execptions,
        # --> should be handled above this level
        # if (xmlSchema.assertValid(xml_etree)):
        if (xmlSchema.validate(xml_etree)):
            return True
        else:
            if self.logger.isEnabledFor(logging.DEBUG):
                self.logger.error('Input failed to Validate')
                log = xmlSchema.error_log
                self.logger.error(log.last_error)
            return False

    # http://lxml.de/xpathxslt.html#xslt
    def xml_transform(self, xml_doc, xslt):
        lxml_transform = etree.XSLT(self.xslt)
        return lxml_transform(xml_doc)

    # Generator Function which processes and returns batches of the input CSV
    # http://pandas.pydata.org/pandas-docs/stable/io.html#io-chunking
    # https://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_csv.html
    #
    # file_reader = pandas.read_csv( ... ,iterator=True)
    def next_file_slice(self, file_reader):
        while True:
            try:
                df_slice = file_reader.get_chunk(self.row_limit)
                if(not self.row_header_in):
                    self.row_header_in = df_slice.columns.values.tolist()
                yield (
                    df_slice.fillna("").values.astype("unicode").tolist(),
                    df_slice.first_valid_index(),
                    df_slice.last_valid_index()
                )
            except StopIteration:
                self.logger.debug('End of input file')
                break

    def write_file_header(self, row_names):
        pd.DataFrame(columns=row_names).to_csv(
            self.fpath_output,
            encoding='utf-8',
            header=True,
            index=False,
        )

    def print_xml(self, etree_obj):
        self.logger.debug('___________________________________________')
        self.logger.debug(etree.tostring(etree_obj, pretty_print=True))



# -*- coding: utf-8 -*-

__all__ = [
    'DEFAULT_RTREE_INDEX_PROPS',
    'generate_index_entries',
    'get_peril_areas',
    'get_peril_areas_index',
    'get_rtree_index',
    'PerilArea',
    'PERIL_ID_FLOOD',
    'PERIL_ID_QUAKE',
    'PERIL_ID_SURGE',
    'PERIL_ID_WIND',
    'PerilPoint'
]

import io
import json
import os
import re
import uuid

import rtree

from rtree.index import (
    Index,
    Property,
)

from rtree.core import RTreeError

from shapely import speedups as shapely_speedups

if shapely_speedups.available:
    shapely_speedups.enable()

from shapely.geometry import (
    box,
    Point,
    MultiPoint,
    Polygon,
)

import six

from .exceptions import OasisException


PERIL_ID_WIND = 1
PERIL_ID_SURGE = 2
PERIL_ID_QUAKE = 3
PERIL_ID_FLOOD = 4

DEFAULT_RTREE_INDEX_PROPS = {
    'buffering_capacity': 10,
    'custom_storage_callbacks': None,
    'custom_storage_callbacks_size': 0,
    'dat_extension': u'dat',
    'dimension': 2,
    'filename': u'',
    'fill_factor': 0.7,
    'idx_extension': u'idx',
    'index_capacity': 100,
    'index_id': None,
    'leaf_capacity': 100,
    'near_minimum_overlap_factor': 32,
    'overwrite': True,
    'pagesize': 4096,
    'point_pool_capacity': 500,
    'region_pool_capacity': 1000,
    'reinsert_factor': 0.3,
    'split_distribution_factor': 0.4,
    'storage': 0,
    'tight_mbr': True,
    'tpr_horizon': 20.0,
    'type': 0,
    'variant': 2,
    'writethrough': False
}


def generate_index_entries(items, objects=None):
    if objects:
        for (key, poly_bounds), obj in zip(items, objects):
            yield key, poly_bounds, obj
    else:
        for key, poly_bounds in items:
            yield key, poly_bounds, None


def get_peril_areas(areas):
    for peril_area_id, coordinates, other_props in areas:
        yield PerilArea(coordinates, peril_area_id=peril_area_id, **other_props)


def get_peril_areas_index(
    areas=None,
    peril_areas=None,
    objects=None,
    properties=None
):
    if not (areas or peril_areas):
        raise OasisException('Either areas or peril areas must be provided')

    items = (
        (pa.id, pa.bounds) for pa in (get_peril_areas(areas) if not peril_areas else peril_areas)
    )

    return get_rtree_index(items, objects=objects, properties=properties)


def get_rtree_index(
    items,
    objects=None,
    properties=None
):
    return (
        Index(generate_index_entries(items, objects=objects), properties=Property(**properties)) if properties
        else Index(generate_index_entries(items, objects=objects))
    )


class PerilArea(Polygon):

    def __init__(self, coords, **kwargs):

        _coords = tuple(c for c in coords if c != (0,0))

        if not _coords:
            raise OasisException('No peril area coordinates')

        if len(_coords) > 1:
            self.multipoint = MultiPoint(_coords)
        elif len(_coords) == 1:
            x, y = _coords[0][0], _coords[0][1]
            r = kwargs.get('area_reg_poly_radius') or 0.0016
            self.multipoint = MultiPoint(
                tuple((x + r*(-1)**i, y + r*(-1)**j) for i in range(2) for j in range(2))
            )

        super(self.__class__, self).__init__(shell=self.multipoint.convex_hull.exterior.coords)
        
        self.coordinates = tuple(self.exterior.coords)

        self.centre = self.centroid.x, self.centroid.y

        self.peril_id = kwargs.get('peril_id')

        self.id = kwargs.get('area_peril_id') or kwargs.get('peril_area_id') or int(uuid.UUID(bytes=os.urandom(16)).hex[:16], 16)

        for k, v in six.iteritems(kwargs):
            setattr(self, k, v)


class PerilPoint(Point):

    def __init__(self, *args, **kwargs):

        super(self.__class__, self).__init__(*args)
        
        self.coordinates = tuple(t for t in self.coords[0])

        self.peril_id = kwargs.get('peril_id')

        self.id = kwargs.get('area_peril_id') or kwargs.get('peril_area_id') or int(uuid.UUID(bytes=os.urandom(16)).hex[:16], 16)

        for k, v in six.iteritems(kwargs):
            setattr(self, k, v)



"""
    Provides a simple class to build Oasis API clients.
"""
import logging

from requests import RequestException
from six.moves import urllib

from ..model_execution.files import TAR_FILE
from ..model_execution.bin import create_binary_tar_file, create_binary_files, check_inputs_directory, \
    cleanup_bin_directory, check_conversion_tools
from ..utils.exceptions import OasisException

__all__ = [
    'OasisAPIClient'
]


# Python 2 standard imports
import os
import io
import time

# Python 3rd party imports
import requests

from requests_toolbelt.multipart.encoder import MultipartEncoder

# Oasis imports
from ..utils.log import oasis_log
from ..utils.status import STATUS_PENDING, STATUS_SUCCESS, STATUS_FAILURE


class OasisAPIClient(object):
    """
    Class for interacting with the oasis api server

    :param oasis_api_url: The root URL for the API. This should
        include the scheme and port (eg. http://localhost:8001)
    :type oasis_api_url: str

    :param logger: The logger to use for message logging. If None
        the root logger is used.
    :type logger: Logger
    """
    #: The chunk size to use when streaming data from the server
    DOWNLOAD_CHUCK_SIZE_IN_BYTES = 1024

    def __init__(self, oasis_api_url, logger=None):
        """
        Construct the client.
        """

        self._oasis_api_url = oasis_api_url
        self._logger = logger or logging.getLogger()

    def build_uri(self, path):
        """
        Builds the uri for the requested resource

        :param path: The path to the resource
        :type path: str

        :return: The fully qualified uri for the resource
        """
        return urllib.parse.urljoin(self._oasis_api_url, path)

    @oasis_log
    def upload_inputs_from_directory(self, directory, bin_directory=None, do_il=False, do_build=False, do_clean=False):
        """
        Upload the CSV files from a specified directory.

        :param directory: the directory containing the CSVs.
        :type directory: str

        :param bin_directory: the directory to build the binary files in, if not set ``directory`` is used
        :type bin_directory: str

        :param do_il: if True, require files for insured loss (IL) calculation.
        :type do_il: bool

        :param do_build: if True, the input tar will be built
        :type do_build: bool

        :param do_clean: if True, remove the tar and bin files
        :type do_clean: bool

        :return: The location of the uploaded inputs.
        """
        bin_directory = bin_directory or directory

        try:
            if do_build:
                check_inputs_directory(directory, do_il=do_il)
                check_conversion_tools(do_il=do_il)
                create_binary_files(directory, bin_directory, do_il=do_il)
                create_binary_tar_file(bin_directory)

            self._logger.debug("Uploading inputs")
            inputs_tar_to_upload = os.path.join(bin_directory, TAR_FILE)

            with io.open(inputs_tar_to_upload, 'rb') as f:
                inputs_multipart_data = MultipartEncoder(
                    fields={
                        'file': (TAR_FILE, f, 'text/plain')
                    }
                )

                response = requests.post(
                    self.build_uri('/exposure'),
                    data=inputs_multipart_data,
                    headers={'Content-Type': inputs_multipart_data.content_type}
                )

            if not response.ok:
                self._logger.error(
                    "POST {} failed: {}, {}".format(response.request.url, str(response.status_code), str(response.json))
                )
                raise OasisException("Failed to save exposure.")

            exposure_location = response.json()['exposures'][0]['location']
            self._logger.debug("Uploaded exposure. Location: " + exposure_location)

            # Return the location of the uploaded inputs
            return exposure_location
        finally:
            # Tidy up
            if do_clean:
                cleanup_bin_directory(bin_directory)

    @oasis_log
    def run_analysis(self, analysis_settings_json, input_location):
        """
        Starts an analysis running. Calling code will need to poll for
        completion or failure, and handle results.

        :param analysis_setting_json: The analysis settings encoded as JSON.
        :type analysis_setting_json: str

        :param input_location: The location of the inputs resource.
        :type input_location: str

        :return: The location of analysis status, to poll.
        """
        response = requests.post(
            self.build_uri("/analysis/" + input_location),
            json=analysis_settings_json,
        )

        if not response.ok:
            self._logger.error("POST {} failed: {}".format(response.request.url, str(response.status_code)))
            raise OasisException("Failed to start analysis")

        analysis_status_location = response.json()['location']
        self._logger.info("Analysis started")

        return analysis_status_location

    def get_analysis_status(self, analysis_status_location):
        """
        Fetches the analysis status for the requested analysis.

        :param analysis_status_location: The location the analysis status.
            This is the value returned from ``run_analysis``
        :type analysis_status_location: str

        :raises OasisException: If the analysis fails.
        :raises OasisException: If the http request is not successful

        :return: A 2 tuple of the status and output location. If the
            status is pending the output location is empty.
        """
        response = requests.get(self.build_uri('/analysis_status/' + analysis_status_location))
        if response.status_code != 200:
            raise OasisException("GET analysis status failed: {}".format(response.status_code))

        self._logger.debug("Response: {}".format(response.json()))
        status = response.json()['status']
        self._logger.debug("Analysis status: " + status)

        if status == STATUS_FAILURE:
            error_message = "Analysis failed: {}".format(response.json()['message'])
            self._logger.error(error_message)
            raise OasisException(error_message)
        elif status == STATUS_SUCCESS:
            return status, response.json()['outputs_location']
        else:
            return status, ''

    @oasis_log
    def run_analysis_and_poll(self, analysis_settings_json, input_location, outputs_directory, analysis_poll_interval=5):
        """
        Run an analysis to completion or failure. resources on the server are
        cleaned up once the analysis is complete.

        :param analysis_setting_json: The analysis settings encoded as JSON.
        :type analysis_settings_json: str

        :param input_location: The local directory containing the inputs resources.
        :type input_location: str

        :param outputs_directory: The local directory to save the outputs to.
        :type outputs_directory: str

        :param analysis_poll_interval: The interval time to wait between status checks
        :type analysis_poll_interval: int

        :raises OasisException: If the analysis fails.
        :raises OasisException: If a http request is not successful
        """
        analysis_status_location = self.run_analysis(analysis_settings_json, input_location)

        self._logger.info("Analysis started")

        status, outputs_location = self.get_analysis_status(analysis_status_location)
        # Poll until the analysis has completed successfully.
        # If the analysis fails then an exception will be raised.
        while status != STATUS_SUCCESS:
            time.sleep(analysis_poll_interval)
            status, outputs_location = self.get_analysis_status(analysis_status_location)

        self._logger.debug("Analysis completed")

        self._logger.debug("Downloading outputs")
        outputs_file = os.path.join(outputs_directory, outputs_location + ".tar.gz")
        self.download_outputs(outputs_location, outputs_file)
        self._logger.debug("Downloaded outputs")

        # cleanup
        self.delete_exposure(input_location)
        self.delete_outputs(outputs_location)

    def delete_resource(self, path):
        """
        Cleans up a resource on the server

        :param path: The path to the resource to delete
        :type path: str
        """
        response = requests.delete(self.build_uri(path))
        if response.status_code != 200:
            self._logger.warning("DELETE {} failed: {}".format(response.request.url, response.status_code))
        else:
            self._logger.info("Deleted {}".format(response.request.url))

    @oasis_log
    def delete_exposure(self, input_location):
        """
        Cleans up an exposure on the server

        :param input_location: The input location for the analysis
        :type input_location: str
        """
        self._logger.info("Deleting exposure")
        self.delete_resource('/exposure/' + input_location)

    @oasis_log
    def delete_outputs(self, outputs_location):
        """
        Cleans up ouptut files on the server

        :param output_location: The output location for the analysis
        :type output_location: str
        """
        self._logger.info("Deleting outputs")
        self.delete_resource('/outputs/' + outputs_location)

    def download_resource(self, path, localfile):
        """
        Streams a resource from the server to a local file

        :param path: The path of the resource to download
        :type path: str

        :param localfile: The path to the local file to download the
            resource to.
        :type localfile: str
        """
        if os.path.exists(localfile):
            error_message = 'Local file alreday exists: {}'.format(localfile)
            self._logger.error(error_message)
            raise OasisException(error_message)

        response = requests.get(self.build_uri(path), stream=True)
        if not response.ok:
            exception_message = 'GET {} failed: {}'.format(response.request.url, response.status_code)
            self._logger.error(exception_message)
            raise OasisException(exception_message)

        with io.open(localfile, 'wb') as f:
            for chunk in response.iter_content(chunk_size=self.DOWNLOAD_CHUCK_SIZE_IN_BYTES):
                f.write(chunk)

    @oasis_log
    def download_exposure(self, exposure_location, localfile):
        """
        Download exposure data to a specified local file.

        :param exposure_location: The location of the exposure resource.
        :type exposure_location: str

        :param localfile: The path to the local file to download the
            exposure to.
        :type localfile: str
        """
        self.download_resource('/exposure/' + exposure_location, localfile)

    @oasis_log
    def download_outputs(self, outputs_location, localfile):
        """
        Download outputs data to a specified local file.

        :param outputs_location: The location of the ouptuts resource.
        :type outputs_location: str

        :param localfile: The path to the local file to download the
            exposure to.
        :type localfile: str
        """
        self.download_resource('/outputs/' + outputs_location, localfile)

    @oasis_log
    def health_check(self, poll_attempts=1, retry_delay=5):
        """
        Checks the health of the server.

        :param poll_attempts: The maximum number of checks to make
        :type poll_attempts: int

        :param retry_delay: The amount of time to wait between retry attempts
        :type retry_delay: int

        :return: True If the server is healthy, otherwise False
        """
        for attempt in range(poll_attempts):
            try:
                if attempt > 0:
                    time.sleep(retry_delay)

                resp = requests.get('{}/healthcheck'.format(self._oasis_api_url))
                if resp.status_code == 200:
                    return True
            except RequestException:
                pass
        else:
            self._logger.error(
                'Could not connect to the api server after {} attempts. Check it is running and try again later.'.format(
                    poll_attempts,
                )
            )
            return False



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

__all__ = [
    'OasisBaseLookup',
    'OasisBaseKeysLookup',
    'OasisPerilAndVulnerabilityLookup',
    'OasisPerilLookup',
    'OasisVulnerabilityLookup',
    'OasisKeysLookupFactory'
]

import csv
import imp
import importlib
import io
import json
import os
import re
import sys
import types
import uuid

from collections import OrderedDict

import pandas as pd
import six

from shapely.geometry import (
    box,
    Point,
)

from shapely import speedups as shapely_speedups

if shapely_speedups.available:
    shapely_speedups.enable()

from rtree.core import RTreeError

from six import StringIO

from ..utils.data import get_dataframe
from ..utils.exceptions import OasisException
from ..utils.log import oasis_log
from ..utils.peril import (
    DEFAULT_RTREE_INDEX_PROPS,
    get_peril_areas_index,
    PerilArea,
    PerilPoint,
)
from ..utils.status import (
    KEYS_STATUS_FAIL,
    KEYS_STATUS_NOMATCH,
    KEYS_STATUS_SUCCESS,
)


UNKNOWN_ID = -1

class OasisBaseLookup(object):

    @oasis_log()
    def __init__(self, config=None, config_json=None, config_fp=None):
        self.config = config or self.get_config(config_json=config_json, config_fp=config_fp) or {}

        mc = self.config.get('model') or {}
        self.supplier_id = mc.get('supplier_id')
        self.model_id = mc.get('model_id')
        self.model_version = mc.get('model_version')

    @oasis_log()
    def get_config(self, config_json=None, config_fp=None):
        if config_json:
            return json.loads(config_json)

        if config_fp:
            _config_fp = os.path.abspath(config_fp) if not os.path.isabs(config_fp) else config_fp
            with io.open(_config_fp, 'r', encoding='utf-8') as f:
                return json.load(f)

    def lookup(self, loc, **kwargs):
        """
        Lookup for an individual location item, which could be a dict or a
        Pandas series object.
        """
        pass

    @oasis_log()
    def bulk_lookup(self, locs, **kwargs):
        """
        Bulk vulnerability lookup for a generator, list, tuple or dict of
        location items, which can be dicts or Pandas series objects.

        Generates results using ``yield``.
        """
        _locs = (
            enumerate(locs) if isinstance(locs, tuple) or isinstance(locs, list) or isinstance(locs, types.GeneratorType)
            else six.iteritems(locs)
        )
        for _, loc in _locs:
            yield self.lookup(loc)


class OasisBaseKeysLookup(object):  # pragma: no cover
    """
    A base class / interface that serves a template for model-specific keys
    lookup classes.
    """
    @oasis_log()
    def __init__(
        self,
        keys_data_directory=None,
        supplier=None,
        model_name=None,
        model_version=None
    ):
        """
        Class constructor
        """
        if keys_data_directory is not None:
            self.keys_data_directory = keys_data_directory
        else:
            self.keys_data_directory = os.path.join(os.sep, 'var', 'oasis', 'keys_data')

        self.supplier = supplier
        self.model_name = model_name
        self.model_version = model_version

    @oasis_log()
    def process_locations(self, loc_df):
        """
        Process location rows - passed in as a pandas dataframe.
        """
        pass

    def _get_location_record(self, raw_loc_item):
        """
        Returns a dict of standard location keys and values based on
        a raw location item, which is a row in a Pandas dataframe.
        """
        pass

    def _get_area_peril_id(self, record):
        """
        Get the area peril ID for a particular location record.
        """
        return UNKNOWN_ID, "Not implemented"

    def _get_vulnerability_id(self, record):
        """
        Get the vulnerability ID for a particular location record.
        """
        return UNKNOWN_ID, "Not implemented"

    @oasis_log()
    def _get_area_peril_ids(self, loc_data, include_context=True):
        """
        Generates area peril IDs in two modes - if include_context is
        True (default) it will generate location records/rows including
        the area peril IDs, otherwise it will generate pairs of location
        IDs and the corresponding area peril IDs.
        """
        pass

    @oasis_log()
    def _get_vulnerability_ids(self, loc_data, include_context=True):
        """
        Generates vulnerability IDs in two modes - if include_context is
        True (default) it will generate location records/rows including
        the area peril IDs, otherwise it will generate pairs of location
        IDs and the corresponding vulnerability IDs.
        """
        pass

    def _get_lookup_success(self, ap_id, vul_id):
        """
        Determine the status of the keys lookup.
        """
        if ap_id == UNKNOWN_ID or vul_id == UNKNOWN_ID:
            return KEYS_STATUS_NOMATCH
        return KEYS_STATUS_SUCCESS


class OasisKeysLookupFactory(object):
    """
    A factory class to load and run keys lookup services for different
    models/suppliers.
    """
    @classmethod
    def get_model_info(cls, model_version_file_path):
        """
        Get model information from the model version file.
        """
        with io.open(model_version_file_path, 'r', encoding='utf-8') as f:
            return next(csv.DictReader(
                f, fieldnames=['supplier_id', 'model_id', 'model_version_id']
            ))

    @classmethod
    def get_lookup_package(cls, lookup_package_path):
        """
        Returns the lookup service parent package (called `keys_server` and
        located in `src` in the model keys server Git repository or in
        `var/www/oasis` in the keys server Docker container) from the given
        path.
        """
        if lookup_package_path and not os.path.isabs(lookup_package_path):
            lookup_package_path = os.path.abspath(lookup_package_path)

        parent_dir = os.path.abspath(os.path.dirname(lookup_package_path))
        package_name = re.sub(r'\.py$', '', os.path.basename(lookup_package_path))

        sys.path.insert(0, parent_dir)
        lookup_package = importlib.import_module(package_name)
        imp.reload(lookup_package)
        sys.path.pop(0)

        return lookup_package

    @classmethod
    def get_lookup_class_instance(cls, lookup_package, keys_data_path, model_info):
        """
        Get the keys lookup class instance.
        """
        klc = getattr(lookup_package, '{}KeysLookup'.format(model_info['model_id']))

        return klc(
            keys_data_directory=keys_data_path,
            supplier=model_info['supplier_id'],
            model_name=model_info['model_id'],
            model_version=model_info['model_version_id']
        )

    @classmethod
    def get_model_exposures(cls, model_exposures=None, model_exposures_file_path=None):
        """
        Get the model exposures/location file data as a pandas dataframe given
        either the path of the model exposures file or the string contents of
        such a file.
        """
        if model_exposures_file_path:
            loc_df = pd.read_csv(os.path.abspath(model_exposures_file_path), float_precision='high')
        elif model_exposures:
            loc_df = pd.read_csv(StringIO(model_exposures), float_precision='high')
        else:
            raise OasisException('Either model_exposures_file_path or model_exposures must be specified')

        loc_df = loc_df.where(loc_df.notnull(), None)
        loc_df.columns = loc_df.columns.str.lower()

        return loc_df

    @classmethod
    def write_oasis_keys_file(cls, records, output_file_path):
        """
        Writes an Oasis keys file from an iterable of keys records.
        """
        heading_row = OrderedDict([
            ('id', 'LocID'),
            ('peril_id', 'PerilID'),
            ('coverage', 'CoverageID'),
            ('area_peril_id', 'AreaPerilID'),
            ('vulnerability_id', 'VulnerabilityID'),
        ])

        pd.DataFrame(
            columns=heading_row.keys(),
            data=[heading_row] + records,
        ).to_csv(
            output_file_path,
            index=False,
            encoding='utf-8',
            header=False,
        )

        return output_file_path, len(records)

    @classmethod
    def write_oasis_keys_errors_file(cls, records, output_file_path):
        """
        Writes an Oasis keys errors file from an iterable of keys records.
        """
        heading_row = OrderedDict([
            ('id', 'LocID'),
            ('peril_id', 'PerilID'),
            ('coverage', 'CoverageID'),
            ('message', 'Message'),
        ])

        pd.DataFrame(
            columns=heading_row.keys(),
            data=[heading_row] + records,
        ).to_csv(
            output_file_path,
            index=False,
            encoding='utf-8',
            header=False,
        )

        return output_file_path, len(records)

    @classmethod
    def write_json_keys_file(cls, records, output_file_path):
        """
        Writes the keys records as a simple list to file.
        """
        with io.open(output_file_path, 'w', encoding='utf-8') as f:
            f.write(u'{}'.format(json.dumps(records, sort_keys=True, indent=4, ensure_ascii=False)))

            return output_file_path, len(records)

    @classmethod
    def create(
        cls,
        model_keys_data_path=None,
        model_version_file_path=None,
        lookup_package_path=None,
    ):
        """
        Creates a keys lookup class instance for the given model and supplier -
        local file paths are required for the model keys data folder, the model
        version file and the Git repository for the model keys server. Returns a
        pair ``(model_info, klc)``, where ``model_info`` is a dictionary holding
        model information from the model version file and `klc` is the lookup
        service class instance for the model.
        """
        for p in [model_keys_data_path, model_version_file_path, lookup_package_path]:
            p = os.path.abspath(p) if p and not os.path.isabs(p) else p

        model_info = cls.get_model_info(model_version_file_path)
        lookup_package = cls.get_lookup_package(lookup_package_path)
        klc = cls.get_lookup_class_instance(lookup_package, model_keys_data_path, model_info)
        return model_info, klc

    @classmethod
    def get_keys(
        cls,
        lookup=None,
        model_exposures=None,
        model_exposures_file_path=None,
        success_only=True
    ):
        """
        Generates keys keys records (JSON) for the given model and supplier -
        requires an instance of the lookup service (which can be created using
        the `create` method in this factory class), and either the model
        location file path or the string contents of such a file.

        The optional keyword argument ``success_only`` indicates whether only
        records with successful lookups should be returned (default), or all
        records.
        """
        if not (model_exposures or model_exposures_file_path):
            raise OasisException('No model exposures provided')

        model_loc_df = cls.get_model_exposures(
            model_exposures_file_path=model_exposures_file_path,
            model_exposures=model_exposures
        )

        for record in lookup.process_locations(model_loc_df):
            if success_only:
                if record['status'].lower() == KEYS_STATUS_SUCCESS:
                    yield record
            else:
                yield record

    @classmethod
    def save_keys(
        cls,
        lookup=None,
        keys_file_path=None,
        keys_errors_file_path=None,
        keys_format='oasis',
        model_exposures=None,
        model_exposures_file_path=None,
    ):
        """
        Writes a keys file, and optionally a keys error file, for the keys
        generated by the lookup service for the given model, supplier and
        exposure sfile - requires a lookup service instance (which can be
        created using the `create` method in this factory class), the path of
        the model location file, the path of the keys file, and the format of
        the output file which can be an Oasis keys file (``oasis``) or a
        simple listing of the records to file (``json``).

        The optional keyword argument ``keys_error_file_path`` if present
        indicates that all keys records, whether for locations with successful
        or unsuccessful lookups, will be generated and written to separate
        files. A keys record with a successful lookup will have a `status`
        field whose value will be `success`, otherwise the record will have
        a `status` field value of `failure` or `nomatch`.

        If ``keys_errors_file_path`` is not present then the method returns a
        pair ``(p, n)`` where ``p`` is the keys file path and ``n`` is the
        number of "successful" keys records written to the keys file, otherwise
        it returns a quadruple ``(p1, n1, p2, n2)`` where ``p1`` is the keys
        file path, ``n1`` is the number of "successful" keys records written to
        the keys file, ``p2`` is the keys errors file path and ``n2`` is the
        number of "unsuccessful" keys records written to keys errors file.
        """
        if not (model_exposures or model_exposures_file_path):
            raise OasisException('No model exposures or model exposures file path provided')

        keys_file_path, keys_errors_file_path, model_exposures_file_path = map(
            lambda p: os.path.abspath(p) if p and not os.path.isabs(p) else p,
            [keys_file_path, keys_errors_file_path, model_exposures_file_path]
        )

        keys = cls.get_keys(
            lookup=lookup,
            model_exposures=model_exposures,
            model_exposures_file_path=model_exposures_file_path,
            success_only=(True if not keys_errors_file_path else False)
        )

        successes = []
        nonsuccesses = []
        for k in keys:
            successes.append(k) if k['status'] == KEYS_STATUS_SUCCESS else nonsuccesses.append(k)

        if keys_format == 'json':
            if keys_error_file_path:
                fp1, n1 = cls.write_json_keys_file(successes, keys_file_path)
                fp2, n2 = cls.write_json_keys_file(nonsuccesses, keys_errors_file_path)
                return fp1, n1, fp2, n2
            return cls.write_json_keys_file(successes, keys_file_path)
        elif keys_format == 'oasis':
            if keys_errors_file_path:
                fp1, n1 = cls.write_oasis_keys_file(successes, keys_file_path)
                fp2, n2 = cls.write_oasis_keys_errors_file(nonsuccesses, keys_errors_file_path)
                return fp1, n1, fp2, n2
            return cls.write_oasis_keys_file(successes, keys_file_path)
        else:
            raise OasisException("Unrecognised keys file output format - valid formats are 'oasis' or 'json'")


class OasisPerilAndVulnerabilityLookup(OasisBaseLookup):
    """
    Combined peril and vulnerability lookup
    """
    @oasis_log()
    def __init__(
        self,
        config=None,
        config_json=None,
        config_fp=None,
        areas=None,
        peril_id=None,
        peril_areas=None,
        peril_areas_index=None,
        peril_areas_index_props=None,
        loc_to_global_areas_boundary_min_distance=0,
        vulnerabilities=None
    ):
        super(self.__class__, self).__init__(
            config=config,
            config_json=config_json,
            config_fp=config_fp
        )

        self.peril_lookup = OasisPerilLookup(
            config=self.config,
            areas=areas,
            peril_id=peril_id,
            peril_areas=peril_areas,
            peril_areas_index=peril_areas_index,
            peril_areas_index_props=peril_areas_index_props,
            loc_to_global_areas_boundary_min_distance=loc_to_global_areas_boundary_min_distance
        )

        self.peril_id = peril_id or self.config['peril'].get('peril_id') or self.peril_lookup.peril_id

        self.peril_area_id_key = self.config['peril']['peril_area_id_col'].lower()

        self.vulnerability_id_key = self.config['vulnerability']['vulnerability_id_col'].lower()

        self.vulnerability_lookup = OasisVulnerabilityLookup(
            config=self.config,
            vulnerabilities=vulnerabilities
        )

    def lookup(self, loc, loc_id_col='id'):

        loc_id = loc.get(loc_id_col) or int(uuid.UUID(bytes=os.urandom(16)).hex[:16], 16)

        pa_lookup = self.peril_lookup.lookup(loc)
        past = pa_lookup['status']
        pamsg = pa_lookup['message']
        
        vln_lookup = self.vulnerability_lookup.lookup(loc)
        vlnst = vln_lookup['status']
        vlnmsg = vln_lookup['message']
        
        # Could optionally call the status lookup method, but it is always
        # better to avoid outside function calls in a `for` loop if possible
        status = (
            KEYS_STATUS_SUCCESS if past == vlnst == KEYS_STATUS_SUCCESS
            else (KEYS_STATUS_FAIL if (past == KEYS_STATUS_FAIL or vlnst == KEYS_STATUS_FAIL) else KEYS_STATUS_NOMATCH)
        )
        
        message = '{}; {}'.format(pamsg, vlnmsg)

        return {
            'id': loc_id,
            'peril_id': self.peril_id,
            self.peril_area_id_key: pa_lookup['peril_area_id'],
            self.vulnerability_id_key: vln_lookup['vulnerability_id'],
            'status': status,
            'message': message
        }


class OasisPerilLookup(OasisBaseLookup):
    """
    Single peril, single coverage type, lon/lat point-area poly lookup using
    an Rtree index to store peril areas - index entries are
    #
    #    (peril area ID, peril area bounds)
    #
    # pairs. Areas must be represented as polygons with vertices which are
    # lon/lat coordinates, and be passed in to the constructor as a list,
    # tuple, or generator of triples of the form
    #
    #    (peril area ID, polygon lon/lat vertices, dict with optional properties)
    #
    # An optional distance measure ``loc_to_global_areas_boundary_min_distance`` can be passed
    # that defines how far, in an abstract unit, a given lon/lat location can be
    # from the boundary of the polygon containing all the individual peril area
    # polygons in order to be assigned an peril area ID. By default this distance
    # is 0, which means any lon/lat location outside the polygon containing all
    # peril area polygons will not be assigned a peril area ID.
    """

    @oasis_log()
    def __init__(
        self,
        config=None,
        config_json=None,
        config_fp=None,
        areas=None,
        peril_id=None,
        peril_areas=None,
        peril_areas_index=None,
        peril_areas_index_props=None,
        loc_to_global_areas_boundary_min_distance=0
    ):
        super(self.__class__, self).__init__(config=config, config_json=config_json, config_fp=config_fp)

        self.peril_id = peril_id or (self.config['peril'].get('peril_id') if self.config.get('peril') else None)

        if areas or peril_areas or self.config.get('peril'):
            self.areas = self.get_areas(areas=areas)

            self.peril_areas = self.get_peril_areas(areas=self.areas, peril_areas=peril_areas)

            self.peril_areas_dict = {pa.id: pa for _, pa in enumerate(self.peril_areas)}

            self.peril_areas_index, self.peril_areas_index_props = self.get_peril_areas_index(
                peril_areas=self.peril_areas,
                peril_areas_index=peril_areas_index,
                peril_areas_index_props=peril_areas_index_props
            )

            self.peril_areas_boundary = box(*self.peril_areas_index.bounds, ccw=False)

            _centroid = self.peril_areas_boundary.centroid
            self.peril_areas_centre = _centroid.x, _centroid.y

            self.loc_to_global_areas_boundary_min_distance = loc_to_global_areas_boundary_min_distance or self.config['peril']['loc_to_global_areas_boundary_min_distance']

        if self.config.get('locations'):
            self.loc_coords_x_col = self.config['locations'].get('coords_x_col') or 'lon'
            self.loc_coords_y_col = self.config['locations'].get('coords_y_col') or 'lat'
            self.loc_coords_type = self.config['locations'].get('coords_type') or 'lonlat'
            self.loc_x_bounds = tuple(self.config['locations'].get('coords_x_bounds')) or (-180, 180,)
            self.loc_y_bounds = tuple(self.config['locations'].get('coords_y_bounds')) or (-90, 90,)

    @oasis_log()
    def get_areas(self, areas=None):

        if not self.config:
            raise OasisException(
                'No lookup configuration provided or set - use `get_config` '
                'on this instance to set it and provide either an actual '
                'model config dict (use `model_config` argument), or a model '
                'config JSON string (use `model_config_json` argument, or a '
                'model config JSON file path (use `model_config_fp` argument)'
            )

        if areas:
            return tuple(areas) if not isinstance(areas, tuple) else areas

        peril_config = self.config.get('peril')

        if not peril_config:
            raise OasisException('No peril config set in the lookup config')

        src_fp = peril_config.get('file_path')

        if not src_fp:
            raise OasisException(
                'No areas file path provided in the lookup config'
            )

        if not os.path.isabs(src_fp):
            src_fp = os.path.abspath(src_fp)

        src_type = str.lower(peril_config.get('file_type') or '') or 'csv'

        float_precision = 'high' if peril_config.get('float_precision_high') else None

        non_na_cols = tuple(col.lower() for col in peril_config['non_na_cols']) if peril_config.get('non_na_cols') else ()

        peril_area_id_col = str.lower(peril_config.get('peril_area_id_col') or '') or 'area_peril_id'
        col_dtypes = {peril_area_id_col: int} if peril_config.get('col_dtypes') == "infer" else {}

        sort_col = peril_config.get('sort_col') or peril_area_id_col
        sort_ascending = peril_config.get('sort_ascending')

        areas_df = get_dataframe(
            src_fp=src_fp,
            src_type=src_type,
            float_precision=float_precision,
            lowercase_cols=True,
            index_col=True,
            non_na_cols=non_na_cols,
            col_dtypes=col_dtypes,
            sort_col=sort_col,
            sort_ascending=sort_ascending
        )

        coords_cols = peril_config.get('area_poly_coords_cols')

        if not coords_cols:
            raise OasisException(
                'A sequence of coordinate pairs must be provided to define areas'
            )

        seq_start = peril_config.get('area_poly_coords_seq_start_idx') or 1

        len_seq = sum(1 if re.match(r'x(\d+)?', k) else 0 for k in six.iterkeys(coords_cols))

        area_reg_poly_radius = peril_config.get('area_reg_poly_radius') or 0.00166

        other_props = {'peril_id': self.peril_id, 'area_reg_poly_radius': area_reg_poly_radius}

        return tuple(
            (
                ar[peril_area_id_col],
                tuple(
                    (ar.get(coords_cols['x{}'.format(i)].lower()) or 0, ar.get(coords_cols['y{}'.format(i)].lower()) or 0)
                    for i in range(seq_start, len_seq + 1)
                ),
                other_props
            ) for _, ar in areas_df.iterrows()
        )

    @oasis_log()
    def get_peril_areas(self, areas=None, peril_areas=None):

        if peril_areas:
            return tuple(peril_areas) if not isinstance(peril_areas, tuple) else peril_areas
        
        return tuple(
            PerilArea(coordinates, peril_area_id=peril_area_id, **other_props)
            for peril_area_id, coordinates, other_props in (areas or self.areas)
        )

    @oasis_log()
    def get_peril_areas_index(self, peril_areas=None, peril_areas_index=None, peril_areas_index_props=None):

        if peril_areas_index:
            return peril_areas_index, peril_areas_index.properties.as_dict()

        _peril_areas_index_props = peril_areas_index_props or self.config['peril']['rtree_index'] or DEFAULT_RTREE_INDEX_PROPS

        _peril_areas_index_props['index_capacity'] = len(peril_areas)
        _peril_areas_index_props['index_pool_capacity'] = len(peril_areas)

        peril_areas = peril_areas or self.peril_areas

        _peril_areas_index = get_peril_areas_index(
            peril_areas=peril_areas,
            objects=None,
            properties=_peril_areas_index_props
        )

        return _peril_areas_index, _peril_areas_index_props

    def lookup(self, loc, loc_id_col='id'):
        """
        Area peril lookup for an individual lon/lat location item, which can be
        provided as a dict or a Pandas series. The data structure should contain
        the keys `lon` or `longitude` for longitude and `lat` or `latitude` for
        latitude.
        """
        peril_id = self.peril_id
        peril_areas = self.peril_areas_dict
        idx = self.peril_areas_index
        boundary = self.peril_areas_boundary
        loc_to_global_areas_boundary_min_distance = self.loc_to_global_areas_boundary_min_distance

        loc_id = loc.get(loc_id_col) or int(uuid.UUID(bytes=os.urandom(16)).hex[:16], 16)

        loc_x_col = self.loc_coords_x_col
        loc_y_col = self.loc_coords_y_col
        loc_x_bounds = self.loc_x_bounds
        loc_y_bounds = self.loc_y_bounds

        x = loc.get(loc_x_col)
        y = loc.get(loc_y_col)

        _pa_lookup = lambda loc_id, x, y, st, paid, pabnds, msg: {
            'id': loc_id,
            loc_x_col: x,
            loc_y_col: y,
            'peril_id': peril_id,
            'status': st,
            'peril_area_id': paid,
            'area_peril_id': paid,
            'area_bounds': pabnds,
            'message': msg
        }

        try:
            x = float(x)
            y = float(y)
            if not ((loc_x_bounds[0] <= x <= loc_x_bounds[1]) and (loc_y_bounds[0] <= y <= loc_y_bounds[1])):
                raise ValueError('{}/{} out of bounds'.format(loc_x_col, loc_y_col))
        except (ValueError, TypeError) as e:
            msg = (
                'Peril area lookup: invalid {}/{} ({}, {}) - {}'
                .format(loc_x_col, loc_y_col, x, y, str(e))
            )
            return _pa_lookup(loc_id, x, y, KEYS_STATUS_FAIL, None, None, msg)

        st = KEYS_STATUS_NOMATCH
        msg = 'No peril area match'
        paid = None
        pabnds = None
        point = x, y

        try:
            paid = list(idx.intersection(point))[0]
        except IndexError:
            try:
                paid = list(idx.nearest(point))[0]
            except IndexError:
                pass
            else:
                p = Point(x, y)
                min_dist = p.distance(boundary)
                if min_dist > loc_to_global_areas_boundary_min_distance:
                    msg = (
                        'Peril area lookup: location is {} units from the '
                        'peril areas global boundary -  the required minimum '
                        'distance is {} units'
                        .format(min_dist, loc_to_global_areas_boundary_min_distance)
                    )
                    return _pa_lookup(loc_id, x, y, KEYS_STATUS_FAIL, None, None, msg)
                pabnds = peril_areas[paid].bounds
                st = KEYS_STATUS_SUCCESS
                msg = (
                    'Successful peril area lookup: {}'.format(paid)
                )
        except RTreeError as e:
            return _pa_lookup(loc_id, x, y, KEYS_STATUS_FAIL, None, None, str(e))
        else:
            pabnds = peril_areas[paid].bounds
            st = KEYS_STATUS_SUCCESS
            msg = 'Successful peril area lookup: {}'.format(paid)

        return _pa_lookup(loc_id, x, y, st, paid, pabnds, msg)


class OasisVulnerabilityLookup(OasisBaseLookup):
    """
    Simple key-value based vulnerability lookup
    """

    @oasis_log()
    def __init__(
        self,
        config=None,
        config_json=None,
        config_fp=None,
        vulnerabilities=None
    ):
        super(self.__class__, self).__init__(config=config, config_json=config_json, config_fp=config_fp)

        if vulnerabilities or self.config.get('vulnerability'):
            self.vulnerabilities = self.get_vulnerabilities(vulnerabilities=vulnerabilities)

    @oasis_log()
    def get_vulnerabilities(self, vulnerabilities=None):
        if not self.config:
            raise OasisException(
                'No lookup configuration provided or set - use `get_config` '
                'on this instance to set it and provide either an actual '
                'model config dict (use `model_config` argument), or a model '
                'config JSON string (use `model_config_json` argument, or a '
                'model config JSON file path (use `model_config_fp` argument)'
            )

        vuln_config = self.config.get('vulnerability')

        if not vuln_config:
            raise OasisException('No vulnerability config set in the lookup config')

        if not vuln_config.get('cols'):
            raise OasisException(
                'The vulnerability file column names must be set in the '
                'vulnerability section of the lookup config'
            )

        cols = tuple(col.lower() for col in vuln_config['cols'])

        if not vuln_config.get('key_cols'):
            raise OasisException(
                'The vulnerability file key column names must be set in the '
                'vulnerability section of the lookup config'
            )

        key_cols = tuple(col.lower() for col in vuln_config['key_cols'])

        vuln_id_col = self.config['vulnerability'].get('vulnerability_id_col') or 'vulnerability_id'

        def _vuln_dict(vuln_enum, key_cols, vuln_id_col):
            return (
                {v[key_cols[0]]:(v.get(vuln_id_col) or v.get('vulnerability_id')) for _, v in vuln_enum} if len(key_cols) == 1
                else {tuple(v[key_cols[i]] for i in range(len(key_cols))):(v.get(vuln_id_col) or v.get('vulnerability_id')) for _, v in vuln_enum}
            )

        if vulnerabilities:
            return _vuln_dict(enumerate(vulnerabilities), key_cols)

        src_fp = vuln_config.get('file_path')

        if not src_fp:
            raise OasisException(
                'No vulnerabilities file path provided in the lookup config'
            )

        if not os.path.isabs(src_fp):
            src_fp = os.path.abspath(src_fp)

        src_type = str.lower(vuln_config.get('file_type') or '') or 'csv'

        float_precision = 'high' if vuln_config.get('float_precision_high') else None

        non_na_cols = tuple(col.lower() for col in vuln_config['non_na_cols']) if vuln_config.get('non_na_cols') else ()

        col_dtypes = {vuln_id_col.lower(): int} if vuln_config.get('col_dtypes') == "infer" else {}

        sort_col = vuln_config.get('sort_col') or vuln_id_col
        sort_ascending = vuln_config.get('sort_ascending')

        vuln_df = get_dataframe(
            src_fp=src_fp,
            src_type=src_type,
            float_precision=float_precision,
            lowercase_cols=True,
            index_col=True,
            non_na_cols=non_na_cols,
            col_dtypes=col_dtypes,
            sort_col=sort_col,
            sort_ascending=sort_ascending
        )

        return _vuln_dict(vuln_df.iterrows(), key_cols, vuln_id_col)

    def lookup(self, loc, loc_id_col='id'):
        """
        Vulnerability lookup for an individual location item, which could be a dict or a
        Pandas series.
        """
        loc_id = loc.get(loc_id_col) or int(uuid.UUID(bytes=os.urandom(16)).hex[:16], 16)
        coverage = loc['coverage']
        class_1 = loc['class_1']

        vln_lookup = lambda loc_id, vlnst, vlnid, vlnmsg: {
            'id': loc_id,
            'status': vlnst,
            'vulnerability_id': vlnid,
            'message': vlnmsg
        }

        try:
            int(coverage) and str(class_1)
        except (TypeError, ValueError):
            return vln_lookup(loc_id, KEYS_STATUS_FAIL, None, 'Vulnerability lookup: invalid location coverage or class 1')

        vlnst = KEYS_STATUS_NOMATCH
        vlnmsg = 'No vulnerability match'
        vlnid = None

        try:
            vlnid = self.vulnerabilities[(loc['coverage'], loc['class_1'])]
        except KeyError:
            pass
        else:
            vlnst = KEYS_STATUS_SUCCESS
            vlnmsg = 'Successful vulnerability lookup: {}'.format(vlnid)

        return vln_lookup(loc_id, vlnst, vlnid, vlnmsg)



import multiprocessing

import subprocess

from ..utils.exceptions import OasisException
from ..utils.log import oasis_log
from .bash import genbash


@oasis_log()
def run(analysis_settings, number_of_processes=-1, filename='run_ktools.sh'):
    if number_of_processes == -1:
        number_of_processes = multiprocessing.cpu_count()

    genbash(number_of_processes, analysis_settings, filename)
    try:
        subprocess.check_call(['bash', filename])
    except subprocess.CalledProcessError as e:
        raise OasisException('Error running ktools: {}'.format(str(e)))



import json
import os
import io

from collections import Counter
from multiprocessing.pool import ThreadPool

import six
from argparsetree import BaseCommand
from backports.tempfile import TemporaryDirectory

from oasislmf.keys.lookup import OasisKeysLookupFactory
from .. import __version__
from ..utils.exceptions import OasisException
from ..utils.conf import replace_in_file
from ..api_client.client import OasisAPIClient
from .base import OasisBaseCommand
from .cleaners import PathCleaner


class TestModelApiCmd(OasisBaseCommand):
    """
    Tests a model api server
    """

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(TestModelApiCmd, self).add_args(parser)

        parser.add_argument(
            'api_server_url', type=str,
            help='Oasis API server URL (including protocol and port), e.g. http://localhost:8001',
        )

        parser.add_argument(
            '-a', '--analysis-settings-file', type=PathCleaner('Analysis settings'), default='./analysis_settings.json',
            help="Analysis settings JSON file (absolute or relative file path)"
        )

        parser.add_argument(
            '-i', '--input-directory', type=PathCleaner('Input directory'), default='./input',
            help="Input data directory (absolute or relative file path)"
        )

        parser.add_argument(
            '-o', '--output-directory', type=PathCleaner('Output directory', preexists=False), default='./output',
            help="Output data directory (absolute or relative file path)"
        )

        parser.add_argument(
            '-n', '--num-analyses', metavar='N', type=int, default='1',
            help='The number of analyses to run.'
        )

        parser.add_argument(
            '-c', '--health-check-attempts', type=int, default=1,
            help='The maximum number of health check attempts.'
        )

    def load_analysis_settings_json(self, analysis_settings_file):
        """
        Loads the analysis settings JSON file into a dict, also creates a separate
        boolean ``do_il`` for doing insured loss calculations.

        :param analysis_settings_file: Path to the analysis settings file to load
        :type analysis_settings_file: str

        :return: a 2 tuple containing the analysis settings dictionary and a bool
            signifying whether il processing is enabled
        """

        with io.open(analysis_settings_file, encoding='utf-8') as f:
            analysis_settings = json.load(f)

        if isinstance(analysis_settings['analysis_settings']['il_output'], six.string_types):
            do_il = analysis_settings['analysis_settings']["il_output"].lower() == 'true'
        else:
            do_il = bool(analysis_settings['analysis_settings']["il_output"])

        return analysis_settings, do_il

    def run_analysis(self, args):
        """
        Invokes model analysis in the client - is used as a worker function for
        threads.

        :param args: a tuple containing (client, input_directory, output_directory,
            analysis_settings, do_il, counter)
        :type args: tuple
        """
        client, input_directory, output_directory, analysis_settings, do_il, counter = args

        try:
            with TemporaryDirectory() as upload_directory:
                input_location = client.upload_inputs_from_directory(input_directory, bin_directory=upload_directory, do_il=do_il, do_build=True)
                client.run_analysis_and_poll(analysis_settings, input_location, output_directory)
                counter['completed'] += 1

        except Exception as e:
            client._logger.exception("Model API test failed: {}".format(str(e)))
            counter['failed'] += 1

    def action(self, args):
        """
        Runs the api checks for the model

        :param args: The arguments from the command line
        :type args: Namespace
        """

        # get client
        client = OasisAPIClient(args.api_server_url, self.logger)

        # Do a server healthcheck
        if not client.health_check(args.health_check_attempts):
            raise OasisException('Health check failed for {}.'.format(args.api_server_url))

        # Make output directory if it does not exist
        if not os.path.exists(args.output_directory):
            os.mkdir(args.output_directory)

        # Load analysis settings JSON file and set up boolean for doing insured
        # loss calculations
        self.logger.info('Loading analysis settings JSON file:')
        analysis_settings, do_il = self.load_analysis_settings_json(args.analysis_settings_file)
        self.logger.info('  OK: analysis_settings={}, do_il={}'.format(analysis_settings, do_il))

        # Prepare and run analyses
        self.logger.info('Running {} analyses'.format(args.num_analyses))
        counter = Counter()

        threads = ThreadPool(processes=args.num_analyses)
        threads.map(
            self.run_analysis,
            ((client, args.input_directory, args.output_directory, analysis_settings, do_il, counter) for i in range(args.num_analyses))
        )

        threads.close()
        threads.join()

        # Summary of run results
        self.logger.info("Finished: {} completed, {} failed".format(counter['completed'], counter['failed']))
        return 0


class GenerateModelTesterDockerFileCmd(OasisBaseCommand):
    """
    Generates a new a model testing dockerfile from the supplied template
    """

    #: The names of the variables to replace in the docker file
    var_names = [
        'CLI_VERSION',
        'OASIS_API_SERVER_URL',
        'SUPPLIER_ID',
        'MODEL_ID',
        'MODEL_VERSION',
        'LOCAL_MODEL_DATA_PATH'
    ]

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(GenerateModelTesterDockerFileCmd, self).add_args(parser)

        parser.add_argument(
            'api_server_url', type=str,
            help='Oasis API server URL (including protocol and port), e.g. http://localhost:8001'
        )

        parser.add_argument(
            '--model-data-directory', type=PathCleaner('Model data path', preexists=False), default='./model_data',
            help='Model data path (relative or absolute path of model version file)'
        )

        parser.add_argument(
            '--model-version-file', type=PathCleaner('Model version file', preexists=False), default=None,
            help='Model version file path (relative or absolute path of model version file), by default <model-data-path>/ModelVersion.csv is used'
        )

    def action(self, args):
        """
        Generates a new a model testing dockerfile from the supplied template

        :param args: The arguments from the command line
        :type args: Namespace
        """
        dockerfile_src = os.path.join(os.path.dirname(__file__), os.path.pardir, '_data', 'Dockerfile.model_api_tester')

        version_file = args.model_version_file or os.path.join(args.model_data_directory, 'ModelVersion.csv')
        version_info = OasisKeysLookupFactory.get_model_info(version_file)

        dockerfile_dst = os.path.join(
            args.model_data_directory,
            'Dockerfile.{}_{}_model_api_tester'.format(version_info['supplier_id'].lower(), version_info['model_id'].lower()),
        )

        replace_in_file(
            dockerfile_src,
            dockerfile_dst,
            ['%{}%'.format(s) for s in self.var_names],
            [
                __version__,
                args.api_server_url,
                version_info['supplier_id'],
                version_info['model_id'],
                version_info['model_version_id'],
                args.model_data_directory,
            ]
        )

        self.logger.info('File created at {}.'.format(dockerfile_dst))
        return 0


class TestCmd(BaseCommand):
    """
    Test models and keys servers
    """

    sub_commands = {
        'model-api': TestModelApiCmd,
        'gen-model-tester-dockerfile': GenerateModelTesterDockerFileCmd,
    }



from datetime import datetime

import pytz

NULL_VALUES = [None, '', 'n/a', 'N/A', 'null', 'Null', 'NULL']


def get_timestamp(thedate=None, fmt='%Y%m%d%H%M%S'):
    """ Get a timestamp """
    d = thedate if thedate else datetime.now()
    return d.strftime(fmt)


def get_utctimestamp(thedate=None, fmt='%Y-%b-%d %H:%M:%S'):
    """
    Returns a UTC timestamp for a given ``datetime.datetime`` in the
    specified string format - the default format is::

        YYYY-MMM-DD HH:MM:SS
    """
    d = thedate.astimezone(pytz.utc) if thedate else datetime.utcnow()
    return d.strftime(fmt)


def to_string(val):
    """
    Converts value to string, with possible additional formatting.
    """
    return '' if val is None else str(val)


def to_int(val):
    """
    Parse a string to int
    """
    return None if val in NULL_VALUES else int(val)


def to_float(val):
    """
    Parse a string to float
    """
    return None if val in NULL_VALUES else float(val)



# Celery task status
STATUS_PENDING = "PENDING"
STATUS_RUNNING = "RUNNING"
STATUS_SUCCESS = "SUCCESS"
STATUS_FAILURE = "FAILURE"

# Keys lookup status strings.
KEYS_STATUS_SUCCESS = "success"
KEYS_STATUS_FAIL = "fail"
KEYS_STATUS_NOMATCH = "nomatch"



from .config import ConfigCmd
from ..utils.exceptions import OasisException
from .model import ModelsCmd
from .base import OasisBaseCommand
from .test import TestCmd
from .version import VersionCmd
from .bin import BinCmd


class RootCmd(OasisBaseCommand):
    """
    Tool for manageing oasislmf models
    """
    sub_commands = {
        'test': TestCmd,
        'version': VersionCmd,
        'bin': BinCmd,
        'model': ModelsCmd,
        'config': ConfigCmd
    }

    def run(self, args=None):
        """
        Runs the command passing in the parsed arguments. If an ``OasisException`` is
        raised the exception is caught, the error is logged and the process exits with
        an error code of 1.

        :param args: The arguments to run the command with. If ``None`` the arguments
            are gathered from the argument parser. This is automatically set when calling
            sub commands and in most cases should not be set for the root command.
        :type args: Namespace

        :return: The status code of the action (0 on success)
        """
        try:
            return super(OasisBaseCommand, self).run(args=args)
        except OasisException as e:
            self.logger.error(str(e))
            return 1



from __future__ import print_function
# -*- coding: utf-8 -*-

__all__ = [
    'OasisExposuresManagerInterface',
    'OasisExposuresManager'
]

import io
import json
import logging
import os
import shutil

import pandas as pd
import six

from interface import Interface, implements

from ..keys.lookup import OasisKeysLookupFactory
from ..utils.exceptions import OasisException
from ..utils.values import get_utctimestamp
from ..models import OasisModel
from .pipeline import OasisFilesPipeline
from .csv_trans import Translator


class OasisExposuresManagerInterface(Interface):  # pragma: no cover
    """
    Interface class form managing a collection of exposures.

    :param oasis_models: A list of Oasis model objects with resources provided in the model objects'
        resources dictionaries.
    :type oasis_models: ``list(OasisModel)``
    """

    def __init__(self, oasis_models=None):
        """
        Class constructor.
        """
        pass

    def add_model(self, oasis_model):
        """
        Adds Oasis model object to the manager and sets up its resources.
        """
        pass

    def delete_model(self, oasis_model):
        """
        Deletes an existing Oasis model object in the manager.
        """
        pass

    def transform_source_to_canonical(self, oasis_model=None, **kwargs):
        """
        Transforms the source exposures/locations for a given ``oasis_model``
        object to a canonical/standard Oasis format.

        All the required resources must be provided either in the model object
        resources dict or the ``kwargs`` dict.

        It is up to the specific implementation of this class of how these
        resources will be named in ``kwargs`` and how they will be used to
        effect the transformation.

        The transform is generic by default, but could be supplier specific if
        required.
        """
        pass

    def transform_canonical_to_model(self, oasis_model=None, **kwargs):
        """
        Transforms the canonical exposures/locations for a given ``oasis_model``
        object to a format understood by Oasis keys lookup
        services.

        All the required resources must be provided either in the model object
        resources dict or the ``kwargs`` dict.

        It is up to the specific implementation of this class of how these
        resources will be named in ``kwargs`` and how they will be used to
        effect the transformation.
        """
        pass

    def get_keys(self, oasis_model=None, **kwargs):
        """
        Generates the Oasis keys and keys error files for a given model object.
        The keys file is a CSV file containing keys lookup information for
        locations with successful lookups, and has the following headers::

            LocID,PerilID,CoverageID,AreaPerilID,VulnerabilityID

        while the keys error file is a CSV file containing keys lookup
        information for locations with unsuccessful lookups (failures,
        no matches) and has the following headers::

            LocID,PerilID,CoverageID,Message

        All the required resources must be provided either in the model object
        resources dict or the ``kwargs`` dict.

        It is up to the specific implementation of this class of how these
        resources will be named in ``kwargs`` and how they will be used to
        effect the transformation.

        A "standard" implementation should use the lookup service factory
        class in ``oasis_utils`` (a submodule of `omdk`) namely

            ``oasis_utils.oasis_keys_lookup_service_utils.KeysLookupServiceFactory``
        """
        pass

    def load_canonical_profile(self, oasis_model=None, **kwargs):
        """
        Loads a JSON string or JSON file representation of the canonical
        exposures profile for a given ``oasis_model``, stores this in the
        model object's resources dict, and returns the object.
        """
        pass

    def generate_items_file(self, oasis_model=None, **kwargs):
        """
        Generates an items file for the given ``oasis_model``.
        """
        pass

    def generate_coverages_file(self, oasis_model=None, **kwargs):
        """
        Generates a coverages file for the given ``oasis_model``.
        """
        pass

    def generate_gulsummaryxref_file(self, oasis_model=None, **kwargs):
        """
        Generates a gulsummaryxref file for the given ``oasis_model``.
        """
        pass

    def generate_oasis_files(self, oasis_model=None, **kwargs):
        """
        For a given ``oasis_model`` generates the standard Oasis files, namely

            ``items.csv``
            ``coverages.csv``
            ``gulsummaryxref.csv``

        All the required resources must be provided either in the model object
        resources dict or the ``kwargs`` dict.

        It is up to the specific implementation of this class of how these
        resources will be named in ``kwargs`` and how they will be used to
        effect the transformation.
        """
        pass

    def create(self, model_supplier_id, model_id, model_version_id, resources=None):
        pass


class OasisExposuresManager(implements(OasisExposuresManagerInterface)):

    def __init__(self, oasis_models=None):
        self.logger = logging.getLogger()

        self.logger.debug('Exposures manager {} initialising'.format(self))

        self.logger.debug('Adding models')
        self._models = {}

        self.add_models(oasis_models)

        self.logger.debug('Exposures manager {} finished initialising'.format(self))

    def add_model(self, oasis_model):
        """
        Adds model to the manager and sets up its resources.
        """
        self._models[oasis_model.key] = oasis_model

        return oasis_model

    def add_models(self, oasis_models):
        """
        Adds a list of Oasis model objects to the manager.
        """
        for model in oasis_models or []:
            self.add_model(model)

    def delete_model(self, oasis_model):
        """
        Deletes an existing Oasis model object in the manager.
        """
        if oasis_model.key in self._models:
            oasis_model.resources['oasis_files_pipeline'].clear()

            del self._models[oasis_model.key]

    def delete_models(self, oasis_models):
        """
        Deletes a list of existing Oasis model objects in the manager.
        """
        for model in oasis_models:
            self.delete_model(model)

    @property
    def keys_lookup_factory(self):
        """
        Keys lookup service factory property - getter only.

            :getter: Gets the current keys lookup service factory instance
        """
        return self._keys_lookup_factory

    @property
    def models(self):
        """
        Model objects dictionary property.

            :getter: Gets the model in the models dict using the optional
                     ``key`` argument. If ``key`` is not given then the dict
                     is returned.

            :setter: Sets the value of the optional ``key`` in the models dict
                     to ``val`` where ``val`` is assumed to be an Oasis model
                     object (``omdk.OasisModel.OasisModel``).

                     If no ``key`` is given then ``val`` is assumed to be a new
                     models dict and is used to replace the existing dict.

            :deleter: Deletes the value of the optional ``key`` in the models
                      dict. If no ``key`` is given then the entire existing
                      dict is cleared.
        """
        return self._models

    @models.setter
    def models(self, val):
        self._models.clear()
        self._models.update(val)

    @models.deleter
    def models(self):
        self._models.clear()

    def transform_source_to_canonical(self, oasis_model=None, **kwargs):
        """
        Transforms the canonical exposures/locations file for a given
        ``oasis_model`` object to a format understood by Oasis keys lookup
        services.

        By default parameters supplied to this function fill be used if present
        otherwise they will be taken from the `oasis_model` resources dictionary
        if the model is supplied.

        :param oasis_model: The model to get keys for
        :type oasis_model: ``OasisModel``

        :param source_exposures_file_path: Path to the source exposures file
        :type source_exposures_file_path: str

        :param source_exposures_validation_file_path: Path to the exposure validation file
        :type source_exposures_validation_file_path: str

        :param source_to_canonical_exposures_transformation_file_path: Path to the exposure transformation file
        :type source_to_canonical_exposures_transformation_file_path: str

        :param canonical_exposures_file_path: Path to the output canonical exposure file
        :type canonical_exposures_file_path: str

        :return: The path to the output canonical exposure file
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        input_file_path = os.path.abspath(kwargs['source_exposures_file_path'])
        validation_file_path = os.path.abspath(kwargs['source_exposures_validation_file_path'])
        transformation_file_path = os.path.abspath(kwargs['source_to_canonical_exposures_transformation_file_path'])
        output_file_path = os.path.abspath(kwargs['canonical_exposures_file_path'])

        translator = Translator(input_file_path, output_file_path, validation_file_path, transformation_file_path, append_row_nums=True)
        translator()

        if oasis_model:
            oasis_model.resources['oasis_files_pipeline'].canonical_exposures_file = output_file_path

        return output_file_path

    def transform_canonical_to_model(self, oasis_model=None, **kwargs):
        """
        Transforms the canonical exposures/locations file for a given
        ``oasis_model`` object to a format understood by Oasis keys lookup
        services.

        By default parameters supplied to this function fill be used if present
        otherwise they will be taken from the `oasis_model` resources dictionary
        if the model is supplied.

        :param oasis_model: The model to get keys for
        :type oasis_model: ``OasisModel``

        :param canonical_exposures_file_path: Path to the canonical exposures file
        :type canonical_exposures_file_path: str

        :param canonical_exposures_validation_file_path: Path to the exposure validation file
        :type canonical_exposures_validation_file_path: str

        :param canonical_to_model_exposures_transformation_file_path: Path to the exposure transformation file
        :type canonical_to_model_exposures_transformation_file_path: str

        :param model_exposures_file_path: Path to the output model exposure file
        :type model_exposures_file_path: str

        :return: The path to the output model exposure file
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        input_file_path = os.path.abspath(kwargs['canonical_exposures_file_path'])
        validation_file_path = os.path.abspath(kwargs['canonical_exposures_validation_file_path'])
        transformation_file_path = os.path.abspath(kwargs['canonical_to_model_exposures_transformation_file_path'])
        output_file_path = os.path.abspath(kwargs['model_exposures_file_path'])

        translator = Translator(input_file_path, output_file_path, validation_file_path, transformation_file_path, append_row_nums=False)
        translator()

        if oasis_model:
            oasis_model.resources['oasis_files_pipeline'].model_exposures_file_path = output_file_path

        return output_file_path

    def load_canonical_profile(
            self,
            oasis_model=None,
            canonical_exposures_profile_json=None,
            canonical_exposures_profile_json_path=None,
            **kwargs):
        """
        Loads a JSON string or JSON file representation of the canonical
        exposures profile for a given ``oasis_model``, stores this in the
        model object's resources dict, and returns the object.
        """
        if oasis_model:
            canonical_exposures_profile_json = canonical_exposures_profile_json or oasis_model.resources.get('canonical_exposures_profile_json')
            canonical_exposures_profile_json_path = canonical_exposures_profile_json_path or oasis_model.resources.get('canonical_exposures_profile_json_path')

        profile = {}
        if canonical_exposures_profile_json:
            profile = json.loads(canonical_exposures_profile_json)
        elif canonical_exposures_profile_json_path:
            with io.open(canonical_exposures_profile_json_path, 'r', encoding='utf-8') as f:
                profile = json.load(f)

        if oasis_model:
            oasis_model.resources['canonical_exposures_profile'] = profile

        return profile

    def get_keys(self, oasis_model=None, model_exposures_file_path=None, lookup=None, keys_file_path=None, keys_errors_file_path=None, **kwargs):
        """
        Generates the Oasis keys and keys error files for a given model object.
        The keys file is a CSV file containing keys lookup information for
        locations with successful lookups, and has the following headers::

            LocID,PerilID,CoverageID,AreaPerilID,VulnerabilityID

        while the keys error file is a CSV file containing keys lookup
        information for locations with unsuccessful lookups (failures,
        no matches) and has the following headers::

            LocID,PerilID,CoverageID,Message

        By default it is assumed that all the resources required for the
        transformation are present in the model object's resources dict,
        if the model is supplied. These can be overridden by providing the
        relevant optional parameters.

        If no model is supplied then the optional paramenters must be
        supplied.

        If the model is supplied the result key file path is stored in the
        models ``file_pipeline.keyfile_path`` property.

        :param oasis_model: The model to get keys for
        :type oasis_model: ``OasisModel``

        :param keys_file_path: Path to the keys file, required if ``oasis_model`` is ``None``
        :type keys_file_path: str

        :param keys_errors_file_path: Path to the keys error file, required if ``oasis_model`` is ``None``
        :type keys_errors_file_path: str

        :param lookup: Path to the keys lookup service to use, required if ``oasis_model`` is ``None``
        :type lookup: str

        :param model_exposures_file_path: Path to the exposures file, required if ``oasis_model`` is ``None``
        :type model_exposures_file_path: str

        :return: The path to the generated keys file
        """
        if oasis_model:
            _model_exposures_file_path = model_exposures_file_path or oasis_model.resources['oasis_files_pipeline'].model_exposures_file_path
            _lookup = lookup or oasis_model.resources.get('lookup')
            _keys_file_path = keys_file_path or oasis_model.resources['oasis_files_pipeline'].keys_file_path
            _keys_errors_file_path = keys_errors_file_path or oasis_model.resources['oasis_files_pipeline'].keys_errors_file_path

        _model_exposures_file_path, _keys_file_path, _keys_errors_file_path = map(
            lambda p: os.path.abspath(p) if p and not os.path.isabs(p) else p,
            [_model_exposures_file_path, _keys_file_path, _keys_errors_file_path]
        )

        _keys_file_path, _, _keys_errors_file_path, _ = OasisKeysLookupFactory().save_keys(
            keys_file_path=_keys_file_path,
            keys_errors_file_path=_keys_errors_file_path,
            lookup=_lookup,
            model_exposures_file_path=_model_exposures_file_path,
        )

        if oasis_model:
            oasis_model.resources['oasis_files_pipeline'].keys_file_path = _keys_file_path
            oasis_model.resources['oasis_files_pipeline'].keys_errors_file_path = _keys_errors_file_path

        return _keys_file_path, _keys_errors_file_path

    def _process_default_kwargs(self, oasis_model=None, **kwargs):
        if oasis_model:
            kwargs.setdefault('source_exposures_file_path', oasis_model.resources.get('source_exposures_file_path'))
            kwargs.setdefault('source_exposures_validation_file_path', oasis_model.resources.get('source_exposures_validation_file_path'))
            kwargs.setdefault('source_to_canonical_exposures_transformation_file_path', oasis_model.resources.get('source_to_canonical_exposures_transformation_file_path'))
            kwargs.setdefault('canonical_exposures_profile', oasis_model.resources.get('canonical_exposures_profile'))
            kwargs.setdefault('canonical_exposures_profile_json', oasis_model.resources.get('canonical_exposures_profile_json'))
            kwargs.setdefault('canonical_exposures_profile_json_path', oasis_model.resources.get('canonical_exposures_profile_json_path'))
            kwargs.setdefault('canonical_exposures_file_path', oasis_model.resources['oasis_files_pipeline'].canonical_exposures_file_path)
            kwargs.setdefault('canonical_exposures_validation_file_path', oasis_model.resources.get('canonical_exposures_validation_file_path'))
            kwargs.setdefault('canonical_to_model_exposures_transformation_file_path', oasis_model.resources.get('canonical_to_model_exposures_transformation_file_path'))
            kwargs.setdefault('model_exposures_file_path', oasis_model.resources['oasis_files_pipeline'].model_exposures_file_path)
            kwargs.setdefault('keys_file_path', oasis_model.resources['oasis_files_pipeline'].keys_file_path)
            kwargs.setdefault('keys_errors_file_path', oasis_model.resources['oasis_files_pipeline'].keys_errors_file_path)
            kwargs.setdefault('items_file_path', oasis_model.resources['oasis_files_pipeline'].items_file_path)
            kwargs.setdefault('coverages_file_path', oasis_model.resources['oasis_files_pipeline'].coverages_file_path)
            kwargs.setdefault('gulsummaryxref_file_path', oasis_model.resources['oasis_files_pipeline'].gulsummaryxref_file_path)

        if not kwargs.get('canonical_exposures_profile'):
            kwargs['canonical_exposures_profile'] = self.load_canonical_profile(
                oasis_model=oasis_model,
                canonical_exposures_profile_json=kwargs.get('canonical_exposures_profile_json'),
                canonical_exposures_profile_json_path=kwargs.get('canonical_exposures_profile_json_path'),
            )

        return kwargs

    def load_master_data_frame(self, canonical_exposures_file_path, keys_file_path, canonical_exposures_profile, **kwargs):
        with io.open(canonical_exposures_file_path, 'r', encoding='utf-8') as cf:
            canexp_df = pd.read_csv(cf, float_precision='high')
            canexp_df = canexp_df.where(canexp_df.notnull(), None)
            canexp_df.columns = canexp_df.columns.str.lower()

        with io.open(keys_file_path, 'r', encoding='utf-8') as kf:
            keys_df = pd.read_csv(kf, float_precision='high')
            keys_df = keys_df.rename(columns={'CoverageID': 'CoverageType'})
            keys_df = keys_df.where(keys_df.notnull(), None)
            keys_df.columns = keys_df.columns.str.lower()

        tiv_fields = tuple(sorted(
            [v for v in six.itervalues(canonical_exposures_profile) if v and v.get('FieldName').lower() == 'tiv']
        ))

        columns = [
            'item_id',
            'coverage_id',
            'tiv',
            'areaperil_id',
            'vulnerability_id',
            'group_id',
            'summary_id',
            'summaryset_id'
        ]
        result = pd.DataFrame(columns=columns, dtype=object)

        for col in columns:
            result[col] = result[col].astype(int) if col != 'tiv' else result[col]

        item_id = 0
        for i in range(len(keys_df)):
            keys_item = keys_df.iloc[i]

            canexp_item = canexp_df[canexp_df['row_id'] == keys_item['locid']]

            if canexp_item.empty:
                raise OasisException(
                    "No matching canonical exposure item found in canonical exposures data frame for keys item {}.".format(keys_item)
                )

            canexp_item = canexp_item.iloc[0]

            tiv_field_matches = tuple(t for t in tiv_fields if t['CoverageTypeID'] == keys_item['coveragetype'])

            for tiv_field in tiv_field_matches:
                tiv_lookup = tiv_field['ProfileElementName'].lower()
                tiv_value = canexp_item[tiv_lookup]
                if tiv_value > 0:
                    item_id += 1
                    result = result.append([{
                        'item_id': item_id,
                        'coverage_id': item_id,
                        'tiv': tiv_value,
                        'areaperil_id': keys_item['areaperilid'],
                        'vulnerability_id': keys_item['vulnerabilityid'],
                        'group_id': item_id,
                        'summary_id': 1,
                        'summaryset_id': 1,
                    }])

        return result

    def _write_csvs(self, columns, data_frame, file_path):
        data_frame.to_csv(
            columns=columns,
            path_or_buf=file_path,
            encoding='utf-8',
            chunksize=1000,
            index=False
        )

    def generate_items_file(self, oasis_model=None, data_frame=None, **kwargs):
        """
        Generates an items file for the given ``oasis_model``.
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        if data_frame is None:
            data_frame = self.load_master_data_frame(**kwargs)

        self._write_csvs(
            ['item_id', 'coverage_id', 'areaperil_id', 'vulnerability_id', 'group_id'],
            data_frame,
            kwargs['items_file_path']
        )

        if oasis_model:
            oasis_model.resources['oasis_files_pipeline'].items_file_path = kwargs['items_file_path']

        return kwargs['items_file_path']

    def generate_coverages_file(self, oasis_model=None, data_frame=None, **kwargs):
        """
        Generates a coverages file for the given ``oasis_model``.
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        if data_frame is None:
            data_frame = self.load_master_data_frame(**kwargs)

        self._write_csvs(
            ['coverage_id', 'tiv'],
            data_frame,
            kwargs['coverages_file_path']
        )

        if oasis_model:
            oasis_model.resources['oasis_files_pipeline'].coverages_file = kwargs.get('coverages_file_path')

        return kwargs.get('coverages_file_path')

    def generate_gulsummaryxref_file(self, oasis_model=None, data_frame=None, **kwargs):
        """
        Generates a gulsummaryxref file for the given ``oasis_model``.
        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)

        if data_frame is None:
            data_frame = self.load_master_data_frame(**kwargs)

        self._write_csvs(
            ['coverage_id', 'summary_id', 'summaryset_id'],
            data_frame,
            kwargs['gulsummaryxref_file_path']
        )

        if oasis_model:
            oasis_model.resources['oasis_files_pipeline'].gulsummaryxref_path = kwargs['gulsummaryxref_file_path']

        return kwargs['gulsummaryxref_file_path']

    def generate_oasis_files(self, oasis_model=None, **kwargs):
        """
        For a given ``oasis_model`` generates the standard Oasis files, namely

            ``items.csv``
            ``coverages.csv``
            ``gulsummaryxref.csv``

        """
        kwargs = self._process_default_kwargs(oasis_model=oasis_model, **kwargs)
        data_frame = self.load_master_data_frame(**kwargs)

        self.generate_items_file(oasis_model=oasis_model, data_frame=data_frame, **kwargs)
        self.generate_coverages_file(oasis_model=oasis_model, data_frame=data_frame, **kwargs)
        self.generate_gulsummaryxref_file(oasis_model=oasis_model, data_frame=data_frame, **kwargs)

        return {
            'items_file_path': kwargs['items_file_path'],
            'coverages_file_path': kwargs['coverages_file_path'],
            'gulsummaryxref_file_path': kwargs['gulsummaryxref_file_path']
        }

    def clear_files_pipeline(self, oasis_model, **kwargs):
        """
        Clears the oasis files pipeline for the given Oasis model object.

        Args:
            ``oasis_model`` (``omdk.models.OasisModel.OasisModel``): The model object.

            ``**kwargs`` (arbitary keyword arguments):

        Returns:
            ``oasis_model`` (``omdk.models.OasisModel.OasisModel``): The model object with its
            Oasis files pipeline cleared.
        """
        oasis_model.resources.get('oasis_files_pipeline').clear()

        return oasis_model

    def start_files_pipeline(self, oasis_model=None, oasis_files_path=None, source_exposures_file_path=None, logger=None):
        """
        Starts the oasis files pipeline for the given Oasis model object,
        which is the generation of the Oasis items, coverages and GUL summary
        files from the source exposures file, canonical exposures profile,
        and associated validation files and transformation files for the
        source and intermediate files (canonical exposures, model exposures).

        Args:
            ``oasis_model`` (``omdk.models.OasisModel.OasisModel``): The model object.

            ``with_model_resources``: (``bool``): Indicates whether to look for
            resources in the model object's resources dictionary or in ``**kwargs``.

            ``**kwargs`` (arbitrary keyword arguments): If ``with_model_resources``
            is ``False`` then the filesystem paths (including filename and extension)
            of the canonical exposures file, the model exposures
        """
        logger = logger or logging.getLogger()
        logger.info('Checking output files directory exists for model')

        if oasis_model and not oasis_files_path:
            oasis_files_path = oasis_model.resources.get('oasis_files_path')

        if not oasis_files_path:
            raise OasisException('No output directory provided.'.format(oasis_model))
        elif not os.path.exists(oasis_files_path):
            raise OasisException('Output directory {} does not exist on the filesystem.'.format(oasis_files_path))

        logger.info('Checking for source exposures file')
        if oasis_model and not source_exposures_file_path:
            source_exposures_file_path = oasis_model.resources['source_exposures_file_path'] if 'source_exposures_file_path' in oasis_model.resources else None
        if not source_exposures_file_path:
            raise OasisException('No source exposures file path provided in arguments or model resources')
        elif not os.path.exists(source_exposures_file_path):
            raise OasisException("Source exposures file path {} does not exist on the filesysem.".format(source_exposures_file_path))

        utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')
        kwargs = self._process_default_kwargs(
            oasis_model=oasis_model,
            source_exposures_file_path=os.path.join(oasis_files_path, os.path.basename(source_exposures_file_path)),
            canonical_exposures_file_path=os.path.join(oasis_files_path, 'canexp-{}.csv'.format(utcnow)),
            model_exposures_file_path=os.path.join(oasis_files_path, 'modexp-{}.csv'.format(utcnow)),
            keys_file_path=os.path.join(oasis_files_path, 'oasiskeys-{}.csv'.format(utcnow)),
            keys_errors_file_path=os.path.join(oasis_files_path, 'oasiskeys-errors-{}.csv'.format(utcnow)),
            items_file_path=os.path.join(oasis_files_path, 'items.csv'),
            coverages_file_path=os.path.join(oasis_files_path, 'coverages.csv'),
            gulsummaryxref_file_path=os.path.join(oasis_files_path, 'gulsummaryxref.csv')
        )

        if not os.path.exists(kwargs['source_exposures_file_path']):
            self.logger.info('Copying source exposures file to input files directory')
            shutil.copy2(source_exposures_file_path, oasis_files_path)

        logger.info('Generating canonical exposures file {canonical_exposures_file_path}'.format(**kwargs))
        self.transform_source_to_canonical(**kwargs)

        logger.info('Generating model exposures file {model_exposures_file_path}'.format(**kwargs))
        self.transform_canonical_to_model(**kwargs)

        logger.info('Generating keys file {keys_file_path} and keys error file {keys_errors_file_path}'.format(**kwargs))
        self.get_keys(oasis_model=oasis_model, **kwargs)

        logger.info('Generating Oasis files for model')
        return self.generate_oasis_files(oasis_model=oasis_model, **kwargs)

    def create(self, model_supplier_id, model_id, model_version_id, resources=None):
        model = OasisModel(
            model_supplier_id,
            model_id,
            model_version_id,
            resources=resources
        )

        # set default resources
        model.resources.setdefault('oasis_files_path', os.path.join('Files', model.key.replace('/', '-')))
        if not os.path.isabs(model.resources.get('oasis_files_path')):
            model.resources['oasis_files_path'] = os.path.abspath(model.resources['oasis_files_path'])

        model.resources.setdefault('oasis_files_pipeline', OasisFilesPipeline(model_key=model.key))
        if not isinstance(model.resources['oasis_files_pipeline'], OasisFilesPipeline):
            raise OasisException(
                'Oasis files pipeline object for model {} is not of type {}'.format(model, OasisFilesPipeline))

        if model.resources.get('canonical_exposures_profile') is None:
            self.load_canonical_profile(oasis_model=model)

        self.add_model(model)

        return model



from __future__ import print_function, division

import glob
import os
import io
import re
import shutil
import sys
import tarfile
from contextlib import contextmanager
from distutils.log import INFO, WARN, ERROR
from tempfile import mkdtemp
from time import sleep

from setuptools import find_packages, setup, Command
from setuptools.command.install import install

try:
    from urllib.request import urlopen
    from urllib.error import URLError
except ImportError:
    from urllib2 import urlopen, URLError

KTOOLS_VERSION = '0_0_392_0'
SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))


def get_readme():
    with io.open(os.path.join(SCRIPT_DIR, 'README.md'), encoding='utf-8') as readme:
        return readme.read()


def get_install_requirements():
    with io.open(os.path.join(SCRIPT_DIR, 'requirements-package.in'), encoding='utf-8') as reqs:
        return reqs.readlines()


def get_version():
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    with io.open(os.path.join(SCRIPT_DIR, 'oasislmf', '__init__.py'), encoding='utf-8') as init_py:
        return re.search('__version__ = [\'"]([^\'"]+)[\'"]', init_py.read()).group(1)


@contextmanager
def temp_dir():
    d = mkdtemp()
    yield d
    shutil.rmtree(d, ignore_errors=True)


version = get_version()
reqs = get_install_requirements()
readme = get_readme()


class PostInstallKtools(install):
    command_name = 'install'
    user_options = install.user_options + [
        ('ktools', None, 'Only install ktools components'),
    ]
    boolean_options = install.boolean_options + ['ktools']

    def __init__(self, *args, **kwargs):
        self.ktools_components = []
        install.__init__(self, *args, **kwargs)

    def run(self):
        self.install_ktools()
        install.run(self)

    def get_outputs(self):
        return install.get_outputs(self) + self.ktools_components

    def fetch_ktools_tar(self, location, attempts=3, timeout=5, cooldown=1):
        self.announce('Retrieving ktools {}'.format(KTOOLS_VERSION), INFO)

        last_error = None
        request = None
        for i in range(attempts):
            try:
                request = urlopen('https://github.com/OasisLMF/ktools/archive/OASIS_{}.tar.gz'.format(KTOOLS_VERSION), timeout=timeout * 1000)
                break
            except URLError as e:
                self.announce('Failed to get ktools tar (attempt {})'.format(i + 1), WARN)
                last_error = e
                sleep(cooldown)
        else:
            self.announce('Failed to get ktools tar after {} attempts'.format(attempts), ERROR)
            if last_error:
                raise last_error

        with open(location, 'wb') as f:
            f.write(request.read())

    def unpack_tar(self, tar_location, extract_location):
        self.announce('Unpacking ktools', INFO)
        with tarfile.open(tar_location) as tar:
            if not os.path.exists(extract_location):
                os.makedirs(extract_location)
            tar.extractall(extract_location)

    def build_ktools(self, extract_location):
        self.announce('Building ktools', INFO)
        build_dir = os.path.join(extract_location, 'ktools-OASIS_{}'.format(KTOOLS_VERSION))

        os.system('cd {build_dir} && ./autogen.sh && ./configure && make && make check'.format(build_dir=build_dir))
        return build_dir

    def add_ktools_to_path(self, build_dir):
        print('Installing ktools')

        if not os.path.exists(self.install_scripts):
            os.makedirs(self.install_scripts)

        for p in glob.glob(os.path.join(build_dir, 'src', '*', '*')):
            split = p.split(os.path.sep)

            # if the file name is the same as the directory we have found a
            # component executable
            if split[-1] == split[-2]:
                component_path = os.path.join(self.install_scripts, split[-1])
                shutil.copy(p, component_path)
                yield component_path

    def install_ktools(self):
        with temp_dir() as d:
            local_tar_path = os.path.join(d, 'ktools.tar.gz')
            local_extract_path = os.path.join(d, 'extracted')

            self.fetch_ktools_tar(local_tar_path)
            self.unpack_tar(local_tar_path, local_extract_path)
            build_dir = self.build_ktools(local_extract_path)
            self.ktools_components = list(self.add_ktools_to_path(build_dir))


try:
    from wheel.bdist_wheel import bdist_wheel

    class BdistWheel(bdist_wheel):
        command_name = 'bdist_wheel'
        user_options = bdist_wheel.user_options

        def finalize_options(self):
            bdist_wheel.finalize_options(self)
            self.root_is_pure = False

        def get_tag(self):
            python, abi, plat = bdist_wheel.get_tag(self)
            python, abi = 'py2.py3', 'none'
            return python, abi, plat

except ImportError:
    BdistWheel = None

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


class Publish(Command):
    command_name = 'publish'
    user_options = [
        ('wheel', None, 'Publish the wheel'),
        ('sdist', None, 'Publish the sdist tar'),
        ('no-clean', None, 'Don\'t clean the build artifacts'),
        ('sign', None, 'Sign the artifacts using GPG')
    ]
    boolean_options = ['wheel', 'sdist']

    def initialize_options(self):
        self.wheel = False
        self.sdist = False
        self.no_clean = False
        self.sign = False

    def finalize_options(self):
        if not (self.wheel or self.sdist):
            self.announce('Either --wheel and/or --sdist must be provided', ERROR)
            sys.exit(1)

    def run(self):
        if os.system('pip freeze | grep twine'):
            self.announce('twine not installed.\nUse `pip install twine`.\nExiting.', WARN)
            sys.exit(1)

        if self.sdist:
            os.system('python setup.py sdist')

        if self.wheel:
            os.system('python setup.py bdist_wheel')

        if self.sign:
            for p in glob.glob('dist/*'):
                os.system('gpg --detach-sign -a {}'.format(p))

        os.system('twine upload dist/*')
        print('You probably want to also tag the version now:')
        print('  git tag -a {v} -m \'version {v}\''.format(v=version))
        print('  git push --tags')

        if not self.no_clean:
            shutil.rmtree('dist')
            shutil.rmtree('build')
            shutil.rmtree('oasislmf.egg-info')


setup(
    name='oasislmf',
    version=version,
    packages=find_packages(exclude=('tests', 'tests.*', 'tests.*.*')),
    include_package_data=True,
    package_data={
        '': [
            'requirements-package.in',
            'LICENSE',
        ],
        'oasislmf/_data/': ['*']
    },
    exclude_package_data={
        '': ['__pycache__', '*.py[co]'],
    },
    scripts=['bin/oasislmf'],
    license='BSD 3-Clause',
    description='Core loss modelling framework.',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/OasisLMF/oasislmf',
    author='Dan Bate,Sandeep Murthy',
    author_email="Dan Bate <dan.bate@wildfish.com>,Sandeep Murthy <sandeep.murthy@oasislmf.org>",
    keywords='oasis lmf loss modeling framework',
    install_requires=reqs,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    cmdclass={
        'install': PostInstallKtools,
        'bdist_wheel': BdistWheel,
        'publish': Publish,
    },
)
